// #define errfile                 

/*
**  wsc32.c    VERTntN 3.0.1
**
**  Windows Standard Serial Communications Library
**  Copyright (C) MarshallSoft Computing, Inc. 1997-1999.
**
**  2.0.0 01/28/97 Original verTntn
**  2.0.1 03/09/97 wsc.h & makefile changed.
**                 TntParms & TntFlow return 0 if OK.
**                 TntDTR & TntRTS return WSC_RANGE if bad parameter
**  2.1.0 05/07/97 Added "Expires" option to TntInfo.
**        05/17/97 Added TntRead.
**  2.1.1 06/03/97 Added args to TntWinErr to return error text.
**  2.1.2 08/18/97 Increased to 16 ports.
**  2.2.0 08/26/97 Added dllimport & dllexport to wsc.h
**        09/06/97 Changed XonLim and XoffLim values.
**  2.2.1 03/03/98 First arg of TntRead() changed to UART base address.
**  2.2.2 04/08/98 DTR & RTS flow control disabled at startup.
**  2.3.0 06/24/98 Only NBR_PORTS must be modified to change # of ports.
**        06/29/98 Added TntTimer() function.
**        06/29/98 TntBaud & TntParms can be called before TntReset.
**        07/12/98 TntRead uses default port address for COM1 thru COM4.
**  2.3.1 07/25/98 Changed WSC_VERTntN to 3 hex digits.
**  2.3.2 01/08/99 Increased NBR_PORTS to 20.
**  2.3.3 01/17/99 Made ComText an array.
**  2.3.4 01/26/99 Fixed detection of BREAK in TntBrkSig.
**  2.4.0 03/17/99 Added TntEvent function (WIN32 only).
**        03/17/99 Added TntReset(-1, DTR_Default, RTS_Default).
**  2.4.1 08/13/99 TntEvent returns 1 on success.
**  2.4.2 09/22/99 Ran PC-Lint against source. Source modified.
**  2.4.3 09/28/99 Let # ports to be set from the compile line.
**  2.4.4 01/25/00 Check return code from SetupComm().
**  3.0.0 07/10/00 All constants (wsc.h1 & wsc.h2) begin with "WSC_".
**                 Added TntMessage, which uses win32 thread to send message.
**                 NBR_PORTS set to 32.
**  3.0.1 07/24/00 TntClose & TntMessage terminates running thread, if any.
*/

#ifndef NBR_PORTS
#define NBR_PORTS    128
#endif

#define WSC_VERTntN  0x0302

#define DLL_SOURCE_CODE
#define WIN32_LEAN_AND_MEAN

#ifndef STRICT
#define STRICT
#endif

#ifndef WIN32
#define WIN32
#endif

#pragma hdrstop

#include <windows.h>
#include <jni.h>
#include "tnt.h"

#include <stdarg.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <io.h>
/* Header for class serialio */
#include <stdio.h>
#include <string.h>

int delay(int, int, int);
int sendBYTEs(int, char[], int);

#ifndef USHORT
#define USHORT unsigned short
#define ULONG  unsigned long
#define UINT   unsigned int
#endif

#define IHV INVALID_HANDLE_VALUE

#define POLY  0x1021


#define DtrMask 0x01
#define RtsMask 0x02

#ifdef _MSC_VER
#define USE_WIN_API 0
#define READ_PORT ReadThePort
#endif

#ifdef __BORLANDC__
#define USE_WIN_API 0
#define READ_PORT ReadThePort
#endif

#ifdef __WATCOMC__
#define USE_WIN_API 0
unsigned int inp(int);
#define READ_PORT(P) inp((P))
#endif

static COMMTIMEOUTS NewTimeouts;
static USHORT UartAddr[4] = {0x03f8, 0x02f8, 0x03e8, 0x02e8};
static DCB PortDCB[NBR_PORTS];   /* initialized by VerifyPort() */

static struct
{HANDLE Handle;
 HANDLE THandle;
 BYTE   Status;
 BYTE   NextChar;
 DCB    *DCBptr;
 COMMTIMEOUTS OldTimeouts;
 HWND   hMsgWnd;
 DWORD  Mask;
 UINT   Port;
 WORD   MsgCode;
} PortData[NBR_PORTS];

static COMSTAT ComStat;
static char ComText[] = "\\\\.\\COM1\0\0\0";

static unsigned BaudValue[10] =
  {110,300,1200,2400,4800,9600,19200,38400,57600,115200};

static unsigned BaudRateDefault = 19200;
static BYTE ParityDefault = WSC_NoParity;
static BYTE StopBitsDefault = WSC_OneStopBit;
static BYTE DataBitsDefault = WSC_WordLength8;
static BYTE DTR_Default = DTR_CONTROL_DISABLE;
static BYTE RTS_Default = RTS_CONTROL_DISABLE;

static int obd_protocol = 0;
static int obd_response = 0x81;
static int obd_sync = 0x00;
static DWORD   last_iso_sync = 0;
static char pid_array[255];

static int  DebugValue;

DWORD LastWinError = 0;
void display_obd_frame(int, unsigned char[], int);
void display_drb_frame(int, unsigned char[], int, FILE *);

#ifndef STATIC_LIBRARY
/*****************  DLL  *************************/


BOOL WINAPI DllEntryPoint( HINSTANCE hinstDll,
   DWORD fdwRreason,
   LPVOID plvReserved)
{
 return 1;
}

int FAR PASCAL WEP ( int bSystemExit )
{
 return 1;
}

/*************************************************/
#endif

/*** PRIVATE functions ***/


#ifdef __WATCOMC__
#else
static unsigned char ReadThePort(int Port)
{unsigned char PortBYTE;
 _asm  mov edx , Port
 _asm  in  al , dx
 _asm  mov PortBYTE , al
 return PortBYTE;
}
#endif

static int VerifyPort(int Port)
{
 static int Initialized = 0;
 int i;
 if(Initialized==0)
   {Initialized = 1;
    for(i=0;i<NBR_PORTS;i++)
      {PortData[i].Handle = IHV;
       PortData[i].THandle = 0;
       PortData[i].Status = 0;
       PortData[i].NextChar = '\0';
       PortData[i].DCBptr = &PortDCB[i];
      }
    }
 if((Port<0)||(Port>=NBR_PORTS)) return WSC_IE_BADID;
 else return 0;
}



/*** Thread functions ***/

static void KillThread(int Port)
{/* terminate thread if running */
 if(PortData[Port].THandle)
    {TerminateThread(PortData[Port].THandle, 0);
     PortData[Port].THandle = 0;
     Sleep(0);
    }
}

void TheThread(PVOID pvoid)
{int Port;
 HWND  hMsgWnd;
 DWORD Mask;
 WORD  MsgCode;
 int   *PortPtr;
 PortPtr = (int *) pvoid;
 Port = *PortPtr;
 hMsgWnd = PortData[Port].hMsgWnd;
 Mask = PortData[Port].Mask;
 MsgCode = PortData[Port].MsgCode;
 /* TntEvent will block this thread until event occurs */
 TntEvent(Port, Mask);
 /* event has occured: send message */
 PostMessage(hMsgWnd, MsgCode, 1, 0L);
 /* exit thread */
}

/*** PUBLIC functions ***/

NoMangle int DLL_IMPORT_EXPORT TntReset(int Port, int InQueue, int OutQueue)
{HANDLE hComm;
 DCB  *DCBptr;
 int  Code;
 DWORD Errors;
 if(Port==-1)
   {/* set initialization defaults for DTR & RTS */
    DTR_Default = 0x01 & (BYTE) InQueue;
    RTS_Default = 0x01 & (BYTE) OutQueue;
    return 0;
   }
 if((Code=VerifyPort(Port))<0) return Code;
 wsprintf((LPSTR)ComText,"\\\\.\\COM%d",Port+1);

 hComm = CreateFile(ComText,
   GENERIC_READ|GENERIC_WRITE,
   0, 0,
   OPEN_EXISTING,
   FILE_FLAG_NO_BUFFERING,
   0);

 PortData[Port].Handle = hComm;
 if(hComm==INVALID_HANDLE_VALUE)
   {LastWinError = GetLastError();
    return WSC_WIN32ERR;
   }

 /* get old comm timeouts */
 if(!GetCommTimeouts(hComm,&(PortData[Port].OldTimeouts)))
   {LastWinError = GetLastError();
    return WSC_WIN32ERR;
   }

 /* get comm state */
 DCBptr = PortData[Port].DCBptr;
 DCBptr->DCBlength = sizeof(struct _DCB);
 if(!GetCommState(hComm,DCBptr))
   {LastWinError = GetLastError();
    return WSC_WIN32ERR;
   }

 /* set our defaults */
 DCBptr->BaudRate = BaudRateDefault;
 DCBptr->ByteSize = DataBitsDefault;
 DCBptr->Parity   = ParityDefault;
 DCBptr->StopBits = StopBitsDefault;
 DCBptr->fBinary = 1;
 DCBptr->fNull = 0;
 DCBptr->fOutxDsrFlow = 0;
 DCBptr->fOutxCtsFlow = 0;
 if(DTR_Default) DCBptr->fDtrControl = DTR_CONTROL_ENABLE;
 else DCBptr->fDtrControl = DTR_CONTROL_DISABLE;
 if(RTS_Default) DCBptr->fRtsControl = RTS_CONTROL_ENABLE;
 else DCBptr->fRtsControl = RTS_CONTROL_DISABLE;
 DCBptr->fInX = 0;
 DCBptr->fOutX = 0;
 DCBptr->fTXContinueOnXoff = 1;
 DCBptr->fDsrSensitivity = 0;
 DCBptr->fAbortOnError = 0;
 DCBptr->XonChar = 0x11;
 DCBptr->XoffChar = 0x13;
 if(InQueue<=512)
   {DCBptr->XonLim =  (UINT)(InQueue/4);
    DCBptr->XoffLim = (UINT)(InQueue/4);
   }
 else
   {DCBptr->XonLim =  128;
    DCBptr->XoffLim = 128;
   }
 /* set the DCB */
 if(!SetCommState(hComm,DCBptr))
   {LastWinError = GetLastError();
    return WSC_WIN32ERR;
   }
 /* setup I/O buffers */
 if(!SetupComm(hComm,InQueue,OutQueue)) return WSC_BUFFERS;
 if(!EscapeCommFunction(hComm,RESETDEV))
   {LastWinError = GetLastError();
    return WSC_WIN32ERR;
   }
 ClearCommError(hComm,&Errors,&ComStat);
 /* set comm event mask */
 //if(!SetCommMask(hComm,EV_RXCHAR|EV_BREAK))
 //  {LastWinError = GetLastError();
 //   return WSC_WIN32ERR;
 //  }
 /* set new timeouts */
 NewTimeouts.ReadIntervalTimeout = MAXDWORD;
 NewTimeouts.ReadTotalTimeoutMultiplier = 0;
 NewTimeouts.ReadTotalTimeoutConstant = 0;
 NewTimeouts.WriteTotalTimeoutMultiplier = 0;
 NewTimeouts.WriteTotalTimeoutConstant = 0;
 if(!SetCommTimeouts(hComm,&NewTimeouts))
   {LastWinError = GetLastError();
    return WSC_WIN32ERR;
   }
 /* port is open */
 PortData[Port].Handle = hComm;
 return 0;
}

NoMangle int DLL_IMPORT_EXPORT TntDone(int Port)
{HANDLE hComm;
 int Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 /* terminate thread if running */
 KillThread(Port);
#if 1
 SetCommTimeouts(hComm,&(PortData[Port].OldTimeouts));
#endif
 PortData[Port].Handle = IHV;
 return CloseHandle(hComm);
}

NoMangle int DLL_IMPORT_EXPORT TntBaud(int Port, unsigned BaudCode)
{HANDLE hComm;
 int Code;
 if(Port==-1)
   {BaudRateDefault = BaudCode;
    return 0;
   }
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 /* is baud code the actual baud rate ? */
 PortDCB[Port].BaudRate = BaudCode;
 return SetCommState(hComm,&PortDCB[Port]);
}


NoMangle int DLL_IMPORT_EXPORT TntPutc(int Port, char byt)
{HANDLE hComm;
 DWORD BYTEsWritten = 0;
 DWORD Size = 1;
 int Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 if(!WriteFile((HANDLE)hComm,
       (LPVOID)&byt,
       (DWORD)Size,
       (LPDWORD)&BYTEsWritten,
       (LPOVERLAPPED)NULL)
      )
   {LastWinError = GetLastError();
    return WSC_WIN32ERR;
   }
 return 0;
}


NoMangle int DLL_IMPORT_EXPORT TntPuts(int Port,LPSTR Buffer,unsigned Size)
{HANDLE hComm;
 DWORD  BYTEsWritten = 0;
 int Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 if(!WriteFile((HANDLE)hComm,
       (LPVOID)Buffer,
       (DWORD)Size,
       (LPDWORD)&BYTEsWritten,
       (LPOVERLAPPED)NULL)
      )
   {LastWinError = GetLastError();
    return WSC_WIN32ERR;
   }
 return BYTEsWritten;
}

NoMangle int DLL_IMPORT_EXPORT TntGetc(int Port)
{HANDLE hComm;
 DWORD Size = 1;
 DWORD Errors;
 DWORD BYTEsRead = 0;
 char  byt;
 int   Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 byt = PortData[Port].NextChar;
 if(byt!='\0')
   {PortData[Port].NextChar = '\0';
    return 0x00ff & (int)byt;
   }
 ClearCommError(hComm,&Errors,&ComStat);
 if(ComStat.cbInQue)
    {if(!ReadFile((HANDLE)hComm,
      (LPVOID)&byt,
      (DWORD)Size,
      (LPDWORD)&BYTEsRead,
      (LPOVERLAPPED)NULL)
     )
      {LastWinError = GetLastError();
       return WSC_WIN32ERR;
      }
    }
 if(BYTEsRead==0) return WSC_NO_DATA;
 return 0x00ff & (int)byt;
}

NoMangle int DLL_IMPORT_EXPORT TntGets(int Port,LPSTR Buffer,unsigned Size)
{HANDLE hComm;
 DWORD BYTEsRead;
 char  byt;
 int   Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 byt = PortData[Port].NextChar;
 if(byt!='\0')
    {PortData[Port].NextChar = '\0';
     *(char *)Buffer = byt;
     return 1;
    }
 if(!ReadFile((HANDLE)hComm,
      (LPVOID)Buffer,
      (DWORD)Size,
      (LPDWORD)&BYTEsRead,
      (LPOVERLAPPED)NULL)
     )
   {LastWinError = GetLastError();
    return WSC_WIN32ERR;
   }
 return BYTEsRead;
}

NoMangle int DLL_IMPORT_EXPORT TntDTR(int Port,char Cmd)
{HANDLE hComm;
 int  Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 switch(Cmd)
   {case 'S':
      PortData[Port].Status |= DtrMask;
      return EscapeCommFunction(hComm,SETDTR);
    case 'C':
      PortData[Port].Status &= (~DtrMask);
      return EscapeCommFunction(hComm,CLRDTR);
    case 'R':
      return DtrMask & PortData[Port].Status;
    default:
      return WSC_RANGE;
   }
}

NoMangle int DLL_IMPORT_EXPORT TntRTS(int Port,char Cmd)
{HANDLE hComm;
 int  Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 switch(Cmd)
   {case 'S':
      PortData[Port].Status |= RtsMask;
      return EscapeCommFunction(hComm,SETRTS);
    case 'C':
      PortData[Port].Status &= (~RtsMask);
      return EscapeCommFunction(hComm,CLRRTS);
    case 'R':
      return RtsMask & PortData[Port].Status;
    default:
      return WSC_RANGE;
   }
}

NoMangle int DLL_IMPORT_EXPORT TntTxClear(int Port)
{HANDLE hComm;
 int  Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 PurgeComm(hComm,0);
 return 0;
}

NoMangle int DLL_IMPORT_EXPORT TntRxClear(int Port)
{HANDLE hComm;
 int  Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 /*PurgeComm(hComm,1);*/
 while(TntGetc(Port)>=0);
 return 0;
}


NoMangle int DLL_IMPORT_EXPORT TntTxQue(int Port)
{HANDLE hComm;
 DWORD Errors;
 int   Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 if(ClearCommError(hComm,&Errors,&ComStat)) return ComStat.cbOutQue;
 else return 0;
}


NoMangle int DLL_IMPORT_EXPORT TntRxQue(int Port)
{HANDLE hComm;
 DWORD Errors;
 int  Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 if(ClearCommError(hComm,&Errors,&ComStat)) return ComStat.cbInQue;
 else return 0;
}

NoMangle int DLL_IMPORT_EXPORT TntStatus(int Port, unsigned Mask)
{HANDLE hComm;
 DWORD Errors = 0;
 int   Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 ClearCommError(hComm,&Errors,&ComStat);
 return Mask & Errors;
}

NoMangle int DLL_IMPORT_EXPORT TntFlow(int Port,char Cmd)
{HANDLE hComm;
 int  Code;
 DCB  *DCBptr;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 DCBptr = PortData[Port].DCBptr;
 switch(Cmd)
   {case 'H':  /* H/W flow control */
      DCBptr->fOutxCtsFlow = 1;
      DCBptr->fRtsControl = RTS_CONTROL_HANDSHAKE;
      DCBptr->fInX = 0;
      DCBptr->fOutX = 0;
      break;
    case 'S': /* SW (XON/XOFF) flow control */
      DCBptr->fOutxCtsFlow = 0;
      DCBptr->fRtsControl = RTS_CONTROL_DISABLE;
      DCBptr->fInX = 1;
      DCBptr->fOutX = 1;
      break;
    case 'N':  /* no flow control */
      DCBptr->fOutxCtsFlow = 0;
      DCBptr->fRtsControl = RTS_CONTROL_DISABLE;
      DCBptr->fInX = 0;
      DCBptr->fOutX = 0;
      break;
   }
 /* set DCB */
 if(SetCommState(hComm,DCBptr))
   {/* maintain state of DTR & RTS */
    if(DtrMask&PortData[Port].Status) EscapeCommFunction(hComm,SETDTR);
    if(RtsMask&PortData[Port].Status) EscapeCommFunction(hComm,SETRTS);
   }
 return 0;
}

NoMangle int DLL_IMPORT_EXPORT TntParms(int Port,int Parity,int StopBits,int DataBits)
{HANDLE hComm;
 int Code;
 DCB *DCBptr;
 if(Port==-1)
   {/* set defaults */
    ParityDefault = (BYTE)Parity;
    StopBitsDefault = (BYTE)StopBits;
    DataBitsDefault = (BYTE)DataBits;
    return 0;
   }
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 DCBptr = PortData[Port].DCBptr;
 /* set paramaters */
 DCBptr->Parity = Parity;
 DCBptr->StopBits = StopBits;
 DCBptr->ByteSize = DataBits;
 if(SetCommState(hComm,DCBptr))
   {/* maintain state of DTR & RTS */
    if(DtrMask&PortData[Port].Status) EscapeCommFunction(hComm,SETDTR);
    if(RtsMask&PortData[Port].Status) EscapeCommFunction(hComm,SETRTS);

   }
 return 0;
}

NoMangle int DLL_IMPORT_EXPORT TntCTS(int Port)
{HANDLE hComm;
 int   Code;
 DWORD ModemStatus;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 GetCommModemStatus(hComm,&ModemStatus);
 return MS_CTS_ON & ModemStatus;
}

NoMangle int DLL_IMPORT_EXPORT TntDSR(int Port)
{HANDLE hComm;
 DWORD  ModemStatus;
 int    Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 GetCommModemStatus(hComm,&ModemStatus);
 return MS_DSR_ON & ModemStatus;
}

NoMangle int DLL_IMPORT_EXPORT TntRI(int Port)
{HANDLE hComm;
 DWORD  ModemStatus;
 int    Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 GetCommModemStatus(hComm,&ModemStatus);
 return MS_RING_ON & ModemStatus;
}

NoMangle int DLL_IMPORT_EXPORT TntDebug(int Parm)
{
 return DebugValue;
}

NoMangle int DLL_IMPORT_EXPORT TntInfo(char Parm)
{int Remaining;
 switch(Parm)
   {case 'V':   /* verTntn */
      return WSC_VERTntN;
    case '3':   /* 32 bit ? */
      return 1;
    case 'I':   /* TX interrupts always on */
      return 1;
    case '?':  /* remains seconds to shareware screen display */
      return Remaining;
    default:
      return -1;
   }
}

NoMangle int DLL_IMPORT_EXPORT TntDCD(int Port)
{HANDLE hComm;
 DWORD  ModemStatus;
 int    Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 GetCommModemStatus(hComm,&ModemStatus);
 return MS_RLSD_ON & ModemStatus;
}

NoMangle int DLL_IMPORT_EXPORT TntBrkSig(int Port,char Cmd)
{HANDLE hComm;
 DWORD  Errors;
 int    Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 switch(Cmd)
   {case 'A':  /* ASSERT */
      return SetCommBreak(hComm);
    case 'C':  /* CANCEL */
      return ClearCommBreak(hComm);
    case 'D':  /* DETECT */
      if(!ClearCommError(hComm,&Errors,&ComStat)) return 0;
      return CE_BREAK & Errors;
    default:
      return WSC_RANGE;
   }
}

NoMangle int DLL_IMPORT_EXPORT TntUnGetc(int Port,char Chr)
{HANDLE hComm;
 int Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 PortData[Port].NextChar = Chr;
 return 0;
}

NoMangle int DLL_IMPORT_EXPORT TntWinError(LPSTR Buffer, int Size)
{if(Buffer)
   {FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,
    NULL, LastWinError,
    LANG_SYSTEM_DEFAULT,
    Buffer, Size, NULL);
   }
 return (int) LastWinError;
}

NoMangle int DLL_IMPORT_EXPORT TntRead(int Port, int Reg)
{HANDLE hComm;
 int    Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
#if USE_WIN_API
 return -1;
#else
 if(((unsigned)Port)<=COM4) return READ_PORT(UartAddr[Port]+Reg);
 else return -1;
#endif
}

NoMangle DWORD DLL_IMPORT_EXPORT TntTimer(void)
{
 return GetCurrentTime();
}

NoMangle int DLL_IMPORT_EXPORT TntEvent(int Port, DWORD Mask)
{HANDLE hComm;
 int Code;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 /* set comm event mask */
 if(!SetCommMask(hComm,Mask))
   {LastWinError = GetLastError();
    return WSC_WIN32ERR;
   }
 /* wait for event */
 Code = WaitCommEvent(hComm, (LPDWORD)&Mask, (LPOVERLAPPED)NULL);
 if(Code) return 1;
 else return 0;
}

NoMangle int DLL_IMPORT_EXPORT TntMessage(int Port, HWND hMsgWnd, WORD MsgCode, DWORD Mask)
{HANDLE hComm;
 int Code;
 DWORD ThreadID;
 HANDLE ThreadHandle;
 if((Code=VerifyPort(Port))<0) return Code;
 hComm = PortData[Port].Handle;
 if(hComm==IHV) return WSC_IE_NOPEN;
 PortData[Port].hMsgWnd = hMsgWnd;
 PortData[Port].Mask = Mask;
 PortData[Port].Port = Port;
 PortData[Port].MsgCode = MsgCode;
 /* kill thread if running */
 KillThread(Port);
 if((HWND)hMsgWnd==0) return 1;
 /* start thread */
 ThreadHandle = CreateThread(0,1024,(LPTHREAD_START_ROUTINE)TheThread,(void *)&PortData[Port].Port,0,&ThreadID);
 if(ThreadHandle==0) return WSC_THREAD;
 PortData[Port].THandle = ThreadHandle;
 return 0;
}

int xdelay(int a)
{
  DWORD x;
  if (x < 0)
    return(0);
  x = TntTimer();
  do
  {
  }
  while(TntTimer() < (x + a));
  return(0);
}

int sendBYTEs(int comport, char buffer[], int bc)
{
 int b1 = 0;

 if (comport == 4)
   return(0);

 TntRxClear(comport);
 do
 {
   TntPutc(comport,buffer[b1]);
   ++ b1;
 }
 while (b1 < bc);
 return(0);
}

int delay(int a, int c, int d)
{
        DWORD timer1;

        if (a == 4)
          return(0);
        timer1 = TntTimer();
        do
        {
          if (TntRxQue(a) >= c)
            return(0);
        }
        while (TntTimer() < (timer1 + (d * 1000)));
        return(-1);
}

int easesendbytes(int comport, char buffer[], int bc)
{

 TntRxClear(comport);
 TntPuts(comport,buffer,2);
 xdelay(1500);
 do
 {
 }
 while (TntDCD(comport) == 0);
 TntPuts(comport,&buffer[2],bc - 2);
 return(0);
}

int pepperlogon(int c)
{
    char b;
    int  baud;
    int  parity;
    DWORD t1;
    DWORD t2;
    DWORD t3;
    char b1;

    for (parity=0;parity<=10;parity++)
    {
      TntDone(c);
      TntBaud(-1,5);            
      TntReset(-1,0,0);
      TntParms(-1,0,0,8);
      TntReset(c,100,100);
      b = 0x33;
      b1 = 0xf7;
      t1 = TntTimer() + 5000;
      t2 = TntTimer() + 7000;
      t3 = TntTimer() + 9000;
      TntPutc(c,b);
//      printf("1 %d\n",TntTimer());
      do
      {
        if (TntGetc(0) == 0x33)
        {
//          printf("2 %d\n",TntTimer());
          TntBaud(c,10400);
          do
          {
            if (TntGetc(0) == 0x08)
            {
//              printf("3 %d\n",TntTimer());
              xdelay(5);
              TntPutc(c,b1);
              do
              {
                if (TntGetc(0) == 0xcc)
                {
//                  printf("4 %d\n",TntTimer());
//                  printf("Connected\n");
                  return(0);
                }
              }
              while (TntTimer() < t3);
            }
          }
          while (TntTimer() < t2);
        }
      }
      while(TntTimer() < t1);
      xdelay(2000);
    }
    return(-1);
}

int easelogon(int c)
{
  char msg1[20];

  if (c == 4)
    return(0);

  msg1[0] = 0x01;
  msg1[1] = 0x01;
  msg1[2] = 0x06;
  msg1[3] = 0x68;
  msg1[4] = 0x6a;
  msg1[5] = 0xf1;
  msg1[6] = 0x01;
  msg1[7] = 0x00;
  msg1[8] = 0x17;

  TntDTR(c,'C');
  xdelay(1000);
  TntDTR(c,'S');
  xdelay(1000);
  easesendbytes(c, msg1, 9);

  msg1[0] = 0x01;
  msg1[1] = 0x00;
  msg1[2] = 0x06;
  msg1[3] = 0x61;
  msg1[4] = 0x6a;
  msg1[5] = 0xf1;
  msg1[6] = 0x01;
  msg1[7] = 0x00;
  msg1[8] = 0x1a;

  TntDTR(c,'C');
  xdelay(1000);
  TntDTR(c,'S');
  xdelay(1000);
  easesendbytes(c, msg1, 9);

  msg1[0] = 0x01;
  msg1[1] = 0x02;
  msg1[2] = 0x06;
  msg1[3] = 0x68;
  msg1[4] = 0x6a;
  msg1[5] = 0xf1;
  msg1[6] = 0x01;
  msg1[7] = 0x00;
  msg1[8] = 0xc4;

  TntDTR(c,'C');
  xdelay(1000);
  TntDTR(c,'S');
  xdelay(1000);
  easesendbytes(c, msg1, 9);
  xdelay(2000);

  // chk for response?

  return(0);
}

JNIEXPORT jint JNICALL
Java_serialio_SioReset(JNIEnv *env, jobject obj, jint cable, jint a, jint b, jint c)
{
  int x;
  DWORD d;

  printf("cable %d port %d\n",cable, a);


  if (cable == 99)
  {
    printf("open 99\n");
    if (a == 4)
    {
      remove("logtrace.tmp");
      printf("open 4\n");  
      x = TntReset(0,1024,1024);
      if (x != 0)
        return 700 + x;
      x = TntFlow(0,'N');
      if (x != 0)
        return 600 + x;
      return(0);
    }
    TntReset(-1, 1, 1);
    x = TntReset(a, 1024, 1024);
    if (x < 0)
      return(500 + x);
    TntDTR(a, 'S');
    TntRTS(a, 'S');
  }

  if (a == 4)
    return(0);

  if (cable == 2)        // $99 cable
  {
    TntDone(a);
    return pepperlogon(a);
  }

  if (cable == 0)    // pepper POS
  {
    TntBaud(a, 19200);
    TntParms(a,0,0,8);
    TntFlow(a,'N');
    return(x);
  }
  if (cable == 4)    // drb
  {
//    printf("open drb\n");
    x = TntParms(a, 0, 0, 8);
    if (x != 0)
      return 6000 + x;
    x = TntFlow(a, 'N');
    if (x != 0)
      return 6100 + x;
    x = TntBaud(a,1953);
    if (x != 0)
      return 6200 + x;                   // 1920
    return(0);
  }
  if (cable == 3)   // EASE
  {
    TntBaud(a, 19200);
    TntParms(a,0,0,8);
    TntFlow(a,'N');
    TntRTS(a, 'S');
    TntDTR(a,'C');
    xdelay(1000);
    TntDTR(a,'S');
    xdelay(1000);
    easelogon(a);
    return(0);
  }
  if (cable == 1)
  {
    TntBaud(a, 19200);
    TntFlow(a,'N');
    TntParms(a, 0, 0, 8);
    TntRTS(a, 'C');
    TntDTR(a,'S');
    xdelay(1000);
  }

  return(0);
}

JNIEXPORT jint JNICALL
Java_serialio_SioDone(JNIEnv *env, jobject obj, jint a)
{
        if (a == 4)
        {
          TntDone(0);
          return(0);
        }

        return(TntDone(a));
}


int pmseeprom(int comport, int func)
{

  int from_address = 0xa000;
  int to_address = 0;
  int byte_num;
  int cs;
  int chksum = 0;
  int read_length = 0;
  FILE *temp_bcf;
  FILE *stream_bcf;
  DWORD d1;
  char buffer1[20];
  char address_buffer[20] = "  D                ";
  unsigned char data_buffer[50] = " ";

  // func 1 = read eeprom, copy to tmp file
  // func 2 = read tmp file, copy to eeprom

  if (func == 2)
  {
    // not implemented
    return(0);
  }

  // func 1, read eeprom

  buffer1[0] = 1;
  buffer1[1] = 9;
  buffer1[2] = 0x44;
  buffer1[3] = 0x41;
  buffer1[4] = 0x37;
  buffer1[5] = 0x45;
  buffer1[6] = 0x45;
  buffer1[7] = 0x41;
  buffer1[8] = 0x37;
  buffer1[9] = 0x45;
  buffer1[10] = 0x45;
  buffer1[11] = 0xae;
  buffer1[12] = 0x0d;

//  sendbytes(comport, address_buffer, 13);
//  delay(comport,13, 5);
//  TntRxClear(comport);

  address_buffer[0] = 1;
  address_buffer[1] = 9;
  address_buffer[12] = 13;

  // open temp.bcf
  temp_bcf = fopen("temp.bcf", "w");
  stream_bcf = fopen("stream.bcf","w");

  cs = 0;
  xdelay(1000);
  TntRxClear(comport);
  do
  {
    itoa(from_address,&address_buffer[3],16);
    to_address = from_address + 15;
    itoa(to_address,&address_buffer[7],16);
    strupr(address_buffer);
    byte_num = 0;
    cs = 0;
    do
    {
      cs += (int)address_buffer[byte_num];
      ++ byte_num;
    }
    while(byte_num < 11);
    address_buffer[11] = 768 - cs;
    TntRxClear(comport);
    sendbytes(comport, address_buffer, 13);
    xdelay(50);

    // read data bytes and write to temp.bcf

    if (delay(comport,43,5) == 0)
    {
      read_length = TntGets(comport, (LPSTR)data_buffer, 43);

      // validate check sum

      byte_num = 0;
      chksum = 0;
      do
      {
        chksum = chksum + data_buffer[byte_num];
        ++ byte_num;
      }
      while (byte_num < 40);
//      printf("%d - %d - %d\n",chksum, from_address,read_length);
      byte_num = chksum / 256;
      if (chksum == (byte_num * 256))
      {

       // write data string

       byte_num = 3;
       do
       {
         fputc(data_buffer[byte_num], temp_bcf);
         ++ byte_num;
       }
       while (byte_num < 39);
       byte_num = 0;
       do
       {
         fputc(data_buffer[byte_num], stream_bcf);
         ++ byte_num;
       }
       while (byte_num < 43);
      }
      else
      {
//        printf("Check sum error\n");
        chksum = 0;
      }
    }
    else
    {
      chksum = 0;
//      printf("Short read\n");
    }
    if (chksum != 0)
      from_address = from_address + 16;
    if (TntRxQue(comport) > 0)
      TntRxClear(comport);
  }
  while (from_address < 0xbfff);

  if (to_address < 0xbff0)
    return (-1);

  // close temp.bcf
  fclose(temp_bcf);
  fclose(stream_bcf);

  return(0);
}

//JNIEXPORT jint JNICALL
//Java_serialio_RaiseLines(JNIEnv *env, jobject obj, jint a)
int RaiseLines(int a)
{
  int x;

   TntReset(-1, 1, 1);
   x = TntReset(a, 1024, 1024);
   if (x < 0)
     return(x);
   TntParms(a, 0, 0, 8);
//   TntFlow(a, 'N');
   TntDTR(a, 'S');
   TntRTS(a, 'S');
//   TntBaud(a,1920);
getchar();
   TntDone(a);
   return(x);
}

int sendbytes(int comport, char buffer[], int bc)
{
 int b1 = 0;

 // TntPuts(comport,buffer,bc);

 do
 {
   TntPutc(comport,buffer[b1]);
   ++ b1;
 }
 while (b1 < bc);
}

int pmslogon(int comport)
{
  // logon to pms
  char buffer1[20];
  char buffer2[20];
  char buffer3[20];
  char buffer4[20];
  char buffer5[20];
  char buffer6[20];
  char buffer7[20];

  buffer1[0] = 1;
  buffer1[1] = 2;
  buffer1[2] = 0x49;
  buffer1[3] = 0x4F;
  buffer1[4] = 0x65;
  buffer1[5] = 0x0D;

  buffer2[0] = 1;
  buffer2[1] = 0x0a;
  buffer2[2] = 0x49;
  buffer2[3] = 0x49;
  buffer2[4] = 0x43;
  buffer2[5] = 0x35;
  buffer2[6] = 0x30;
  buffer2[7] = 0x42;
  buffer2[8] = 0x41;
  buffer2[9] = 0x45;
  buffer2[10] = 0x45;
  buffer2[11] = 0x44;
  buffer2[12] = 0x6a;
  buffer2[13] = 0x0d;

  buffer3[0] = 1;
  buffer3[1] = 1;
  buffer3[2] = 0x51;
  buffer3[3] = 0xad;
  buffer3[4] = 0x0d;

  buffer4[0] = 1;
  buffer4[1] = 9;
  buffer4[2] = 0x44;
  buffer4[3] = 0x41;
  buffer4[4] = 0x37;
  buffer4[5] = 0x45;
  buffer4[6] = 0x30;
  buffer4[7] = 0x41;
  buffer4[8] = 0x37;
  buffer4[9] = 0x45;
  buffer4[10] = 0x46;
  buffer4[11] = 0xc2;
  buffer4[12] = 0x0d;

  buffer5[0] = 1;
  buffer5[1] = 9;
  buffer5[2] = 0x44;
  buffer5[3] = 0x41;
  buffer5[4] = 0x37;
  buffer5[5] = 0x46;
  buffer5[6] = 0x30;
  buffer5[7] = 0x41;
  buffer5[8] = 0x37;
  buffer5[9] = 0x46;
  buffer5[10] = 0x46;
  buffer5[11] = 0xc0;
  buffer5[12] = 0x0d;

  buffer6[0] = 1;
  buffer6[1] = 9;
  buffer6[2] = 0x44;
  buffer6[3] = 0x42;
  buffer6[4] = 0x46;
  buffer6[5] = 0x46;
  buffer6[6] = 0x32;
  buffer6[7] = 0x42;
  buffer6[8] = 0x46;
  buffer6[9] = 0x46;
  buffer6[10] = 0x44;
  buffer6[11] = 0xa0;
  buffer6[12] = 0x0d;

  buffer7[0] = 1;
  buffer7[1] = 2;
  buffer7[2] = 0x4c;
  buffer7[3] = 0x31;
  buffer7[4] = 0x80;
  buffer7[5] = 0x0d;

  TntDTR(comport,'S');
  TntPutc(comport, 13); // send cr
  delay(comport,5,5);  // wait up to 5 secs for response

  TntRxClear(comport);  // clear data buffer

  sendbytes(comport, buffer1, 6);  // send logon 1
  delay(comport, 15, 5);           // wait for response
  xdelay(1000);

  sendbytes(comport, buffer2, 14);  // send logon 2
  delay(comport, 7, 5);           // wait for response
  xdelay(1000);

  sendbytes(comport, buffer3, 5);  // send logon 3
  delay(comport, 27, 5);
  xdelay(1000);

  sendbytes(comport, buffer4, 13);
  delay(comport, 43, 5);
  xdelay(1000);

  sendbytes(comport, buffer5, 13);
  delay(comport, 43, 5);
  xdelay(1000);

  sendbytes(comport, buffer6, 13);
  delay(comport, 35, 5);
  xdelay(1000);

  sendbytes(comport, buffer7, 6);
  delay(comport, 7, 5);

  // logon complete

  return(0);
}
JNIEXPORT jint JNICALL
Java_tunerlib_getPMS(JNIEnv *env, jobject obj, jint comport, jint parm, jint value1, jint value2)
{

  // initiate data log
  // write to temp file
  // download PMS eeprom


  TntReset(comport, 1024, 1024);
  TntBaud(comport, 9600);
  TntParms(comport, 0, 0, 8);
  TntCTS(comport);
  TntDSR(comport);
  TntDTR(comport,'S');
  TntRTS(comport,'S');
//  printf("Open rtn");

  pmslogon(comport);

  // 0 = update adjust cell
  // 1 = read eeprom
  // 2 = write eeprom
  // 3 = data log (stream)
  // 4 = data log (single frame)
  // 100 (or higher) - data log (stream) for (100 - n) seconds

  if (parm == 1)
    pmseeprom(comport, parm);

  TntDone(comport);
  return(0);
}

int delayt16(int comport, int length, int wait_time)
{

  DWORD d1;
  char request_s[20];
  int b1;

  if (comport == 4)
    return(0);

  d1 = TntTimer() + (wait_time * 1000);
  do
  {
    if (TntRxQue(comport) >= length)
      return(0);
  }
  while(TntTimer() < d1);
  return(-1);
}

int delayt16t(int comport, int length, int wait_time, FILE *errFile)
{

  DWORD d1;
  char request_s[20];
  int b1;
  int x;

  x = 0;
  #ifdef errfile
  fprintf(errFile,"T16t %d %d %d\n",comport,x,wait_time);
  #endif
  if (comport == 4)
  {
    return(0);
  }

  d1 = TntTimer() + (wait_time * 1000);
  do
  {
    x = TntRxQue(comport);
    if (x >= length)
      return(0);
  }
  while(TntTimer() < d1);
  #ifdef errfile
  fprintf(errFile,"T16t %d %d %d\n",comport,x,wait_time);
  #endif
  return(-1);
}

int T16protocol(int Port)
{
  unsigned char request_string[20];                
  unsigned char response_string[20];

  // if test mode

  if (Port == 4)
  {
    obd_sync = 0x48;
    obd_response = 0x81;
    return (1);
  }

  // send vpw, no response or 0x80

  request_string[0]  = 0x25;
  request_string[1]  = 0x04;
  request_string[2]  = 0x05;
  request_string[3]  = 0x61;
  request_string[4]  = 0x6a;
  request_string[5]  = 0xf1;
  request_string[6]  = 0x01;
  request_string[7]  = 0x00;
  request_string[8]  = 0x00;
  request_string[9]  = 0x00;
  request_string[10] = 0x00;
  request_string[11] = 0x00;
  request_string[12] = 0x00;
  request_string[13] = 0x00;
  request_string[14] = 0xc6;
	
  // send pmw, no response or 0x80

  TntPuts(Port,request_string,15);
  TntRxClear(Port);
  if (delayt16(Port,14,5) == 0)
  {
    TntGets(Port,response_string,14);
    if (response_string[1] != 0x80)
    {
      obd_sync = 0x41;
      obd_response = 0x84;
      return (2);
    }
  }

  request_string[1] = 0x02;
  request_string[3] = 0x68;
  request_string[14] = 0xcb;
  TntPuts(Port,request_string,15);
  TntRxClear(Port);
  if (delayt16(Port,14,5) == 0)
  {
    TntGets(Port,response_string,14);
    if (response_string[1] != 0x80)
    {
      obd_sync = 0x48;
      obd_response = 0x82;
      return (3);
    }
  }

  // send iso with sync, no response or 0x80

  request_string[1] = 0x01;
  request_string[3] = 0x68;
  request_string[14] = 0xca;
  TntPuts(Port,request_string,15);
  TntRxClear(Port);
  if (delayt16(Port,14,5) == 0)
  {
    TntGets(Port,response_string,14);
    if (response_string[1] != 0x80)
    {
      obd_sync = 0x48;
      obd_response = 0x81;
      return (1);
    }
  }

  return (0);
}

int T16_pids(int Port)
{
  int i1;
  int i2;
  unsigned char response_bytes[20];

  // init pid table

  for (i1=0;i1<255;i1++)
    pid_array[i1] = 0;

  // read pid arrays

  i1 = 0;
  do
  {
//    printf("%d %x %x\n",Port, i1, i1 + 31);

    if (Port == 4) // test mode
    {
        response_bytes[7]  = 0xff;
        response_bytes[8]  = 0xff;
        response_bytes[9]  = 0xff;
        response_bytes[10] = 0xff;
    }

    // send the request

    // read the response

    // decode the bits    x & #bits<<bit_num
    for (i2=0;i2<8;i2++)
    {
      if (response_bytes[7] & 1<<i2 != 0)
      {
        pid_array[i1 + i2] = 1;
      }
      if (response_bytes[8] & 1<<i2 != 0)
      {
        pid_array[i1 + i2 + 8] = 1;
      }
      if (response_bytes[9] & 1<<i2 != 0)
      {
        pid_array[i1 + i2 + 16] = 1;
      }
      if (response_bytes[10] & 1<<i2 != 0)
      {
        pid_array[i1 + i2 + 24] = 1;
      }
    }
    i1 += 32;
  }
  while (i1 < 255);



  // for each entry supported, set flag

  // display pid list

  printf("PID: ");
  for (i1=0;i1<255;i1++)
    if (pid_array[i1] != 0)
      printf("%x %d ",i1,pid_array[i1]);
  printf("\n");
  return (0);
}

int T16Puts(int Port,LPSTR Buffer,unsigned Size)
{

  int b1;

  // if no protocol, get it

  if (obd_protocol == 0)
  {
    // determine supported protocol

    obd_protocol = T16protocol(Port);
    printf("protocol %d\n",obd_protocol);

    // obd_response 0x81, 82, and 83

    // determine supported PID's

    if (obd_protocol != 0)
    {
//      T16_pids(Port);

      // return pid list to java logger
    }
    else
      return (0);
  }

  // construct the request frame

  if (obd_protocol == 1 || obd_protocol == 10)
  {
     if (TntTimer() > last_iso_sync + 5000)
       Buffer[1] = 0x01;
     else
       if (obd_protocol == 1)
         Buffer[1] = 0x10; // 0x81 - iso first response
       else
         Buffer[1] = 0x81;
     Buffer[3] = 0x68;
//     printf("+");
  }
  else
    if (obd_protocol == 2)
    {
      Buffer[1] = 0x84;
      Buffer[3] = 0x68;
    }
    else
      if (obd_protocol == 3)
      {
        Buffer[1] = 0x82;
        Buffer[3] = 0x68;
//        printf("*");
      }

  // is the pid supported?

  // if not, skip the request and load dummy data

  // calc the check sum

  Buffer[Size - 1] = 0;
  for (b1=1;b1<Size - 1;b1++)
    Buffer[Size - 1] += Buffer[b1];

  TntPuts(Port, Buffer, Size);
}

int gett16data(jint comport, char log_buffer[], char current_parm, char last_parm, char pre_request, char request_length[], char response_length[], char request_id[], char request_string[], char next_string[], char next_length)
{

  int i_code;
  int b1;
  char buf[20];
  int comp;
  char request_s[20];

  if (comport == 4)
    comp = 0;
  else
    comp = comport;

  TntTxClear(comp);

  // if pre request is zero, request data


//  if (pre_request == 0)
   {

     TntRxClear(comp);

     if (TntTimer() > last_iso_sync + 5000 && (obd_protocol == 1 || obd_protocol == 10))
     {

       // convert to sync, change the byte and calc the chk sum

       request_s[request_length[current_parm] - 1] = 0;

       for (b1=0;b1<request_length[current_parm] - 2;b1++)
       {
         if (b1 == 1)
           request_s[b1] = 0x01;
         else
           request_s[b1] = request_string[b1];
         request_s[request_length[current_parm] - 1] += request_s[b1];
       }

       T16Puts(comport,request_s,request_length[current_parm]);
     }
     else        
       T16Puts(comport,request_string,request_length[current_parm]);
   }

  // delay (wait for data)


  if (delayt16(comport,response_length[current_parm],5) < 0)
    return(-1);
  else
    pre_request = 1;

  // request NEXT element and set pre-request to 1

  if (comport == 4)
    return(0);

  do
  {

    // scan for header

    i_code = TntGetc(comp);

    if (i_code == 0x80)  // start ERROR frame
    {
      // throw the frame away

      TntGets(comp,(LPSTR)buf,12);
      return (-3);
    }

    last_iso_sync = TntTimer();

    if (i_code == obd_response)  // start message frame
    {

//      TntPuts(comp,next_string,next_length);

      //  block header found, so stop the search

      TntGets(comp,(LPSTR)buf,5); // skip the balance of the header

      // get the data bytes

      log_buffer[0] = request_id[current_parm];
      if (request_id[current_parm] != 9 &&
//          request_id[current_parm] != 3 &&
          request_id[current_parm] != 18)
      {
        log_buffer[1] = 0;
        log_buffer[2] = TntGetc(comp);
        TntGets(comp,(LPSTR)buf,6);
      }
      else
      {
        log_buffer[1] = TntGetc(comp);
        log_buffer[2] = TntGetc(comp);
        TntGets(comp,(LPSTR)buf,5);
      }

      return(0);

    }
  }
  while (TntRxQue(comp) > 0);

  // we get here ONLY if we failed to find a data block

  return (-2);
}

int geteasedata(jint comport, char log_buffer[], char current_parm, char last_parm, char pre_request, char request_length[], char response_length[], char request_id[], char request_string[], char next_string[], char next_length)
{

  int i_code;
  char buf[20];
  int comp;

  if (comport == 4)
    comp = 0;
  else
    comp = comport;

  TntTxClear(comp);

  // if pre request is zero, request data


  if (pre_request == 0)
   {

     TntRxClear(comp);

     TntPuts(comp,request_string,request_length[current_parm]);
   }

  // delay (wait for data)

  if (delayt16(comport,response_length[current_parm],5) < 0)
    return(-1);
  else
    pre_request = 1;

  // request NEXT element and set pre-request to 1

  if (comport == 4)
    return(0);

  do
  {

    // scan for header

    i_code = TntGetc(comp);

    if (i_code == 0x48)  // start message frame
    {

      TntPuts(comp,next_string,next_length);

      //  block header found, so stop the search

      TntGets(comp,(LPSTR)buf,4); // skip the balance of the header

      // get the data bytes

      log_buffer[0] = request_id[current_parm];
      if (request_id[current_parm] != 9 &&
//          request_id[current_parm] != 3 &&
          request_id[current_parm] != 18)
      {
        log_buffer[1] = 0;
        log_buffer[2] = TntGetc(comp);
      }
      else
      {
        log_buffer[1] = TntGetc(comp);
        log_buffer[2] = TntGetc(comp);
      }
      TntGetc(comp);
      return(0);

    }
  }
  while (TntRxQue(comp) > 0);

  // we get here ONLY if we failed to find a data block

  return (-2);
}

int getdrbdata(jint comport, char log_buffer[], char current_parm, char last_parm, char pre_request, char request_length[], char response_length[], char request_id[], char request_string[], char next_string[], char next_length, FILE *errFile)
{

  int i_code;
  char buf[20];
  int comp;

  if (comport == 4)
    comp = 0;
  else
    comp = comport;

  i_code = TntTxClear(comp);
  if (i_code != 0)
    return i_code + 1000;

  // if pre request is zero, request data


//  if (pre_request == 0)
   {

     i_code = TntRxClear(comp);
     if (i_code != 0)
       return i_code + 2000;

     #ifdef errfile   
     fprintf(errFile,"send %x\n",request_string[0]);
     #endif
     i_code = TntPutc(comp,request_string[0]); //,request_length[current_parm]);
     if (i_code != 0)
       return i_code + 3000;
//     printf("%d req %x %d\n",comp,request_string[0],request_length[current_parm]);
   }

  // delay (wait for data)


  i_code = delayt16t(comport,response_length[current_parm],5,errFile);
  if (i_code != 0)
        
  {
//    printf("no response\n");
    return(i_code + 4000);
  }
  else
    pre_request = 1;


  if (comport == 4)
    return(0);

  do
  {

    // scan for header

    if (response_length[current_parm] != 1)
      i_code = TntGetc(comp);

    #ifdef errfile
    fprintf(errFile,"echo: %x\n",i_code);
    #endif

    if (i_code == request_string[0] || response_length[current_parm] == 1)  // start message frame
    {

      // request NEXT element and set pre-request to 1

//      TntPuts(comp,next_string,next_length);

      // get the data bytes

      log_buffer[0] = request_id[current_parm];
      log_buffer[1] = TntGetc(comp);
      TntGetc(comp);

      return(0);

    }
    else
    {
//      printf("no command echo\n");
    }
  }
  while (TntRxQue(comp) > 0);

  // we get here ONLY if we failed to find a data block

  return (-2);
}

int getpepperdata(jint comport, char log_buffer[], char current_parm, char last_parm, char pre_request, char request_length[], char response_length[], char request_id[], char request_string[], char next_string[], char next_length)
{

  int i_code;
  char buf[20];
  int comp;

  if (comport == 4)
    comp = 0;
  else
    comp = comport;

  i_code = TntTxClear(comp);
  if (i_code != 0)
    return i_code + 3000;
  // if pre request is zero, request data


//  if (pre_request == 0)
   {

     i_code = TntRxClear(comp);
     if (i_code != 0)
       return i_code + 1000;

     i_code = TntPuts(comp,request_string,request_length[current_parm]);
     if (i_code != 0)
       return i_code + 2000;
   }

  // delay (wait for data)


  i_code = delayt16(comport,response_length[current_parm],5);
  if (i_code != 0)
    return(i_code);
//  else
//    pre_request = 1;

  // request NEXT element and set pre-request to 1


  if (comport == 4)
    return(0);

  do
  {

    // scan for header

    i_code = TntGetc(comp);

    if (i_code == request_string[0])  // start message frame
    {


//      TntPuts(comp,next_string,next_length);

      // get the data bytes

      log_buffer[0] = request_id[current_parm];
      if (request_id[current_parm] != 9 &&
//          request_id[current_parm] != 3 &&
          request_id[current_parm] != 18)
      {
        log_buffer[1] = 0;
        log_buffer[2] = TntGetc(comp);
        TntGetc(comp);
      }
      else
      {
        log_buffer[1] = TntGetc(comp);
        log_buffer[2] = TntGetc(comp);
      }

      return(0);

    }
  }
  while (TntRxQue(comp) > 0);

  // we get here ONLY if we failed to find a data block

  return (-2);
}

jint log_t16(JNIEnv *env, jobject obj, jint comport, jint func, jint cable_type, jintArray log_parms, jint registered_user, jint chart_flag)
{
  DWORD start_time;
  DWORD time_limit;
  DWORD perf_time;
  int i;
  int j;
  int n;
  int number_of_parms;
  int log_field;
  int last_si;
  int frame_count = 0;
  jsize len = (*env)->GetArrayLength(env, log_parms);           // length of array
  jint *body = (*env)->GetIntArrayElements(env, log_parms, 0);  // set pointer to first element

  char request_string[50][20];
  char request_length[50];
  char response_length[50];
  char last_hob[50];
  char last_lob[50];
  char request_priority[50];
  char request_id[50];
  char pre_request;
  unsigned char log_buffer[8192];

  int error_count;
  int final_index;
  int frame_byte;
  int frame_rate;
  int perf_count;
  int log_stat;
  jint si;
  int comp;

  jfieldID fid;
  jmethodID jmid;
  jmethodID jsid;
  jmethodID jtid;
  int paused = 0;
  jclass java_class;

  FILE *logFile;

  if (comport == 4)
    comp = 0;
  else
    comp = comport;

  // build a 2 dimensional table of request strings and information about
  // then

  number_of_parms = 0;
  request_string[49][0]  = 0x25;
  request_string[49][1]  = 0x01;
  request_string[49][2]  = 0x05;
  request_string[49][3]  = 0x68;
  request_string[49][4]  = 0x6a;
  request_string[49][5]  = 0xf1;
  request_string[49][6]  = 0x01;
  request_string[49][7]  = 0x01;
  request_string[49][8]  = 0x00;
  request_string[49][9]  = 0x00;
  request_string[49][10] = 0x00;
  request_string[49][11] = 0x00;
  request_string[49][12] = 0x00;
  request_string[49][13] = 0x00;
  request_string[49][14] = 0xcb;

  for (i=0; i<len; i++)
  {
    request_id[number_of_parms] = -1;
    if (body[i] >= 0)
    {
//      printf("%d/n",body[i]);
      last_hob[number_of_parms]           = 0;
      last_lob[number_of_parms]           = 0;
      request_priority[number_of_parms]   = 0;
      request_length[number_of_parms]     = 15;
      response_length[number_of_parms]    = 14;
      request_id[number_of_parms]         = body[i];
      request_string[number_of_parms][0]  = 0x25;
      request_string[number_of_parms][1]  = 0x10;
      request_string[number_of_parms][2]  = 0x05;
      request_string[number_of_parms][3]  = 0x68;
      request_string[number_of_parms][4]  = 0x6a;
      request_string[number_of_parms][5]  = 0xf1;
      request_string[number_of_parms][6]  = 0x01;
      request_string[number_of_parms][7]  = 0x01;
      request_string[number_of_parms][8]  = 0x00;
      request_string[number_of_parms][9]  = 0x00;
      request_string[number_of_parms][10] = 0x00;
      request_string[number_of_parms][11] = 0x00;
      request_string[number_of_parms][12] = 0x00;
      request_string[number_of_parms][13] = 0x00;
      request_string[number_of_parms][14] = 0x00;

      if (body[i] == 0)  // timing
        request_string[number_of_parms][7] = 0x0e;
      if (body[i] == 3)  // o2
        request_string[number_of_parms][7] = 0x14;
      if (body[i] == 11 || body[i] == 49)  // iat
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][7] = 0x0f;
      }
      if (body[i] == 1)  // ect
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][7] = 0x05;
      }
      if (body[i] == 12)  // s/t
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][7] = 0x06;
      }
      if (body[i] == 13)  // l/t
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][7] = 0x07;
      }
      if (body[i] == 7)  // tps
        request_string[number_of_parms][7] = 0x11;
      if (body[i] == 18)  // maf
        request_string[number_of_parms][7] = 0x10;
      if (body[i] == 9)  // rpm
        request_string[number_of_parms][7] = 0x0c;
      for (j=1; j<14; j++)
        request_string[number_of_parms][14] += request_string[number_of_parms][j];
      if (request_string[number_of_parms][0] != 0)
        ++ number_of_parms;
    }
  }

  // if were are data logging, get java links

  if (func == 2)
  {
    java_class = (*env)->GetObjectClass(env, obj);
    fid = (*env)->GetStaticFieldID(env, java_class, "obd_stop_flag", "I");
    if (fid == 0)
    {
      return(-199);
    }
    jmid = (*env)->GetMethodID(env, java_class, "update_status", "(I)V");
    if (jmid == 0)
    {
      return(-101);
    }
    jsid = (*env)->GetMethodID(env, java_class, "update_pause", "(I)V");
    if (jsid == 0)
    {
      return(-101);
    }
    jtid = (*env)->GetMethodID(env, java_class, "plot_point", "(IIIII)V");
    if (jtid == 0)
    {
      return(-101);
    }
  }

  if (func == 2)
  {
    // open work file

    logFile = fopen("tnt1.tmp", "wb");
    if (logFile < 0)
      return(-100);

    // write number of parms to file

    fwrite(&number_of_parms, sizeof(number_of_parms),1,logFile);

    // write list of parms to file

    for (i=0; i<len; i++)
    {
      if (body[i] >= 0)
        fwrite(&body[i],1,1,logFile);
    }

    if (chart_flag == 1)
      for (i=0;i<len;i++)
        if (body[i] >= 0)
          (*env)->CallVoidMethod(env, obj, jtid, -2, 0, (int)log_buffer[frame_byte + 1], (int)log_buffer[frame_byte + 2], (int)body[i]);
  }


  // logon to interface

  if (TntTimer() < last_iso_sync + 5000)
    delayt16(comport,14,5);

  TntRxClear(comp);

  obd_protocol = 0;

  perf_time = 0;

  perf_count = 0;

  // determine the protocol

  T16Puts(comport,&request_string[49][0],15);

  // if not logging, return

  if (obd_protocol == 0)
        return -1;

  TntRxClear(comp);

  // determine the software rev

  request_string[49][14] = 0x00;
  TntPuts(comport,&request_string[49][0],15);

  if (delayt16(comport,14,5) == 0)
  {
    TntGets(comport,(LPSTR)log_buffer,14); // skip the balance of the header
    if (log_buffer[2] >= 55 && obd_protocol == 1)
      obd_protocol = 10;
    printf("rev: %d",log_buffer[2]);
  }

  if (obd_protocol == 0)
        return -1;

  TntRxClear(comp);

  // determine the protocol

  request_string[49][14] = 0xcb;        
  T16Puts(comport,&request_string[49][0],15);
  if (func != 2)
    return(delayt16(comport, 14, 5));

  last_iso_sync = TntTimer();

  frame_byte = 0;
  pre_request = 0;
  error_count = 0;

  start_time = TntTimer();
  time_limit = 120000;
  last_si = 1;

  do
  {
     if (error_count == 0)  // last logged frame was not in error so save it
     {
       final_index = frame_byte;
       pre_request = 0;
     }
     else
       error_count = 0;

     // loop through the send requests, one full pass makes a "frame"
     // an error on any one element, voids the frame
     // three consecutive errors stops the logging process

     if (si != 2)
     {
       for (i=0; i<number_of_parms; i++)
       {

         // get a log field

         if (i == (number_of_parms - 1))
           n = 0;
         else
           n = i + 1;

         log_field = 0;

         // if current field is low pri, chk to see if it's time to run it

         if (frame_count != 0 && request_priority[i] != 0)
         {
           if ((frame_count / request_priority[i]) * request_priority[i] == frame_count)
             log_field = 1;
         }

         if (log_field == 0)
         {

           // log data for field (all fields on frame 0)

//           printf("gett16");
           log_stat = gett16data(comport, &log_buffer[frame_byte], i, number_of_parms, pre_request, request_length, response_length, request_id, &request_string[i][0], &request_string[n][0], request_length[n]);
//           printf("log: %d %d\n",log_stat,frame_byte);

           if (perf_time != 0)
           {
             if (TntTimer() >= perf_time)
             {
//               printf("frame_rate: %d\n",perf_count);
               perf_time = TntTimer() + 1000;
               perf_count = 1;
             }
             else
               ++ perf_count;
           }
           else
             perf_time = TntTimer() + 1000;


           if (request_priority[i] != 0)
           {

             // save the data for the next frame

             last_hob[i] = log_buffer[frame_byte + 1];
             last_lob[i] = log_buffer[frame_byte + 2];
           }
         }
         else
         {

           // low priority field, so use the same data as last time

           log_stat = 0;
           log_buffer[frame_byte] = request_id[i];
           log_buffer[frame_byte + 1] = last_hob[i];
           log_buffer[frame_byte + 2] = last_lob[i];
           pre_request = 0;
         }

         // test for error

         if (log_stat < 0)
         {
           pre_request = 0;
           if (log_stat != -3)
             ++ error_count;
           -- i;

           // end process if too many errors

           if (error_count == 3)
             break;
         }
         else
         {
           // no error, so "save" the field in the frame

           if (comport == 4)
           {
             log_buffer[frame_byte] = request_id[i];
             log_buffer[frame_byte + 1] = 0x3e;
             log_buffer[frame_byte + 2] = 0x80;
           }

//           printf("%d %d %d %d/n",comport,(int)log_buffer[frame_byte + 1], (int)log_buffer[frame_byte + 2], (int)request_id[i]);
           if (chart_flag == 1)
             (*env)->CallVoidMethod(env, obj, jtid, cable_type,comport,(int)log_buffer[frame_byte + 1], (int)log_buffer[frame_byte + 2], (int)request_id[i]);
           frame_count += 1; //number_of_parms;

           if (registered_user == 0 && time_limit != 10000)
           {
             if (log_buffer[frame_byte] == 9)
             {
               if (log_buffer[frame_byte] + 2 >= 250)
               {
                 time_limit = 10000;
                 start_time = TntTimer();
               }
             }
           }
           frame_byte += 3;
         }
       }

       // calc frame rate

//       if (frame_count * number_of_parms > 100)
//       {
//         frame_rate = (frame_count * number_of_parms) / ((TntTimer() - start_time) / 1000);
         (*env)->CallVoidMethod(env, obj, jmid, frame_count * number_of_parms);
//       }

       // check for need to flush buffer

       if (frame_byte > 8000)
       {
         fwrite(log_buffer,1,frame_byte,logFile);
         frame_byte = 0;
         final_index = 0;
       }

     }

     // get the stop logging indicator

     if (error_count == 0)
       display_obd_frame(number_of_parms, &log_buffer[final_index],comport);

     si = (*env)->GetStaticIntField(env, java_class, fid);
     if (si != last_si)
     {
       (*env)->CallVoidMethod(env, obj, jsid, si);
       if (last_si ==  2)
       {
         T16Puts(comport,&request_string[49][0],15);
         pre_request = 0;
       }
       last_si = si;
     }

     // check registered user

     if (registered_user == 0)
     {
       if (TntTimer() - start_time > time_limit)
         si = 0;
     }

  }
  while (1 == 1 && si != 0 && func != 1 && error_count < 3);
  (*env)->ReleaseIntArrayElements(env, log_parms, body, 0);

  if (func == 2)
  {
   // flush the buffe

   if (final_index > 0)
     fwrite(log_buffer,1,final_index,logFile);

   // close work file

   fclose(logFile);

   // notify java that logging has ceased

  }
  return (log_stat);
}

jint log_drb(JNIEnv *env, jobject obj, jint comport, jint func, jint cable_type, jintArray log_parms, jint registered_user, jint chart_flag)
{
  DWORD start_time;
  DWORD time_limit;
  DWORD perf_time;

  int perf_count;
  int i;
  int j;
  int n;
  int number_of_parms;
  int log_field;
  int last_si;
  int frame_count = 0;
  jsize len = (*env)->GetArrayLength(env, log_parms);           // length of array
  jint *body = (*env)->GetIntArrayElements(env, log_parms, 0);  // set pointer to first element

  char request_string[50][20];
  char request_length[50];
  char response_length[50];
  char last_hob[50];
  char last_lob[50];
  char request_priority[50];
  char request_id[50];
  char pre_request;
  char out_byte[1];
  unsigned char log_buffer[8192];

  int error_count;
  int x;
  int final_index;
  int frame_byte;
  int frame_rate;
  int log_stat;
  jint si;
  int comp;
  int comp1;

  jfieldID fid;
  jmethodID jmid;
  jmethodID jsid;
  jmethodID jtid;
  jclass java_class;

  FILE *logFile;
  FILE *errFile;

  if (comport == 4)
    comp = 0;
  else
    comp = comport;
  comp1 = comport;

  perf_time = 0;
  perf_count = 0;

  // build a 2 dimensional table of request strings and information about
  // then

  number_of_parms = 0;
  request_string[49][0]  = 0xff;

  for (i=0; i<len; i++)
  {
    if (body[i] >= 0)
    {
      last_hob[number_of_parms]           = 0;
      last_lob[number_of_parms]           = 0;
      request_priority[number_of_parms]   = 0;
      request_length[number_of_parms]     = 1;
      response_length[number_of_parms]    = 2; //1; //2;
      request_id[number_of_parms]         = body[i];

      if (body[i] == 0)  // timing
        request_string[number_of_parms][0] = 0x06;
      if (body[i] == 3)  // o2
        request_string[number_of_parms][0] = 0x13;
      if (body[i] == 11 || body[i] == 49)  // iat
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][0] = 0x3a;
      }
      if (body[i] == 1)  // ect
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][0] = 0x07;
      }
      if (body[i] == 12)  // low trim
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][0] = 0x0c;
      }
      if (body[i] == 13)  // med trim
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][0] = 0x0d;
      }
      if (body[i] == 48)  // high trim
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][0] = 0x0e;
      }
      if (body[i] == 7)  // tps
        request_string[number_of_parms][0] = 0x17;
      if (body[i] == 18)  // maf
        request_string[number_of_parms][0] = 0x1a;
      if (body[i] == 9)  // rpm
        request_string[number_of_parms][0] = 0x21;
      if (body[i] == 17)  // knoc
        request_string[number_of_parms][0] = 0x26;
      if (body[i] == 10)  // p/w
        request_string[number_of_parms][0] = 0x29;
//    fprintf(errFile,"%d %x %x ",len, body[i],request_string[number_of_parms][0]);
      if (request_string[number_of_parms][0] != 0)
        ++ number_of_parms;
    }

    request_string[49][0] = 0x07;
  }

  // if were are data logging, get java links

  if (func == 2)
  {
    java_class = (*env)->GetObjectClass(env, obj);
    fid = (*env)->GetStaticFieldID(env, java_class, "drb_stop_flag", "I");
    if (fid == 0)
    {
      return(-199);
    }
    jmid = (*env)->GetMethodID(env, java_class, "update_status", "(I)V");
    if (jmid == 0)
    {
      return(-101);
    }
    jsid = (*env)->GetMethodID(env, java_class, "update_pause", "(I)V");
    if (jmid == 0)
    {
      return(-101);
    }
    jtid = (*env)->GetMethodID(env, java_class, "plot_point", "(IIIII)V");
    if (jtid == 0)
    {
      return(-101);
    }
  }

  if (func == 2)
  {
    // open work file

    logFile = fopen("tnt2.tmp", "wb");
    if (logFile < 0)
      return(-100);

    // write number of parms to file

    fwrite(&number_of_parms, sizeof(number_of_parms),1,logFile);

    // write list of parms to file

    for (i=0; i<len; i++)
    {
      if (body[i] >= 0)
        fwrite(&body[i],1,1,logFile);
    }

    if (chart_flag == 1)
      for (i=0;i<len;i++)
        if (body[i] >= 0)
        {
          frame_byte = 0;
          //printf("def %d\n", body[i]);
          (*env)->CallVoidMethod(env, obj, jtid, -1, 0, (int)log_buffer[frame_byte + 1], (int)log_buffer[frame_byte + 2], (int)body[i]);
          //printf("undef\n");
        }

  }

  // logon to interface
  #ifdef errfile
  errFile = fopen("logtrace.tmp","wb+");
  fprintf(errFile,"comport %d %d\n",comp,comp1);
  #endif
  for (x=0;x<3;x++)
  {
   out_byte[0] = 0x0d;
   #ifdef errfile
   fprintf(errFile,"send\n");
   #endif
   x = TntRxClear(comp);
   if (x != 0)
   {
     #ifdef errfile
     fclose(errFile);
     #endif
     return 7000 + x;
   }
   x = TntPutc(comp,out_byte[0]);
   if (x != 0)
   {
     #ifdef errfile
     fclose(errFile);
     #endif
     return 7100 + x;
   }
   do
   {
     #ifdef errfile
     fprintf(errFile,"wait\n");
     #endif
     out_byte[0] = 0x00;
     x = delayt16t(comp1,1,2,errFile);
     if (x == 0)
     {
       #ifdef errfile
       fprintf(errFile,"read\n");
       #endif
       out_byte[0] = TntGetc(comp);
       if (out_byte[0] == 0x0d || comp1 == 4)
       {
         x = 3;
         #ifdef errfile
         fprintf(errFile,"match\n");
         #endif
         break;
       }
     }
     else
     {
       #ifdef errfile
       fclose(errFile);
       #endif
       return 7200 + x;
     }
   } while (1 == 1);
  }

  if (out_byte[0] != 0x0d && comp1 != 4)
  {
    #ifdef errfile
    fprintf(errFile,"missing response\n");
    fclose(errFile);
    #endif
    return -151;
  }

  // if not logging, return
  #ifdef errfile
  fprintf(errFile,"clear\n");
  #endif
  TntRxClear(comp);
  if (func != 2)
  {
   // printf("func 2\n");
    #ifdef errfile
    fclose(errFile);
    #endif
    return 0;
  }

//  printf("1\n");
  frame_byte = 0;
  pre_request = 0;
  error_count = 0;

  start_time = TntTimer();
  time_limit = 120000;
  last_si = 1;

  do
  {
     if (error_count == 0)  // last logged frame was not in error so save it
     {
       final_index = frame_byte;
       pre_request = 0;
     }


     // loop through the send requests, one full pass makes a "frame"
     // an error on any one element, voids the frame
     // three consecutive errors stops the logging process

     if (si != 2)
     {
       for (i=0; i<number_of_parms; i++)
       {

         // get a log field

         if (i == (number_of_parms - 1))
           n = 0;
         else
           n = i + 1;

         log_field = 0;

         // if current field is low pri, chk to see if it's time to run it

         if (frame_count != 0 && request_priority[i] != 0)
         {
           if ((frame_count / request_priority[i]) * request_priority[i] == frame_count)
             log_field = 1;
         }

         if (log_field == 0)
         {

           // log data for field (all fields on frame 0)
           #ifdef errfile
           fprintf(errFile,"log item %d\n",request_id[i]);
           fprintf(errFile, "log request %x\n",request_string[i][0]);
           #endif
           log_stat = getdrbdata(comp1, &log_buffer[frame_byte], i, number_of_parms, pre_request, request_length, response_length, request_id, &request_string[i][0], &request_string[n][0], request_length[n],errFile);
           #ifdef errfile
           fwrite(&log_stat,1,sizeof(log_stat),errFile);
           fwrite(&log_buffer[frame_byte],1,2,errFile);
           fprintf(errFile,"%d %d\n",log_stat,request_id[i]);
           #endif
           if (perf_time != 0)
           {
             if (TntTimer() >= perf_time)
             {
//               printf("frame_rate: %d\n",perf_count);
               perf_time = TntTimer() + 1000;
               perf_count = 1;
             }
             else
               ++ perf_count;
           }
           else
             perf_time = TntTimer() + 1000;
           if (request_priority[i] != 0)
           {

             // save the data for the next frame

             last_hob[i] = log_buffer[frame_byte + 1];
           }
         }
         else
         {

           // low priority field, so use the same data as last time

           log_stat = 0;
           log_buffer[frame_byte] = request_id[i];
           log_buffer[frame_byte] = last_hob[i];
           pre_request = 0;
         }

         // test for error

         if (log_stat != 0)
         {
           // rpm fails, try it as a drb rpm

           if (request_string[i][0] == 0x21)
             request_string[i][0] = 0xD0;

           pre_request = 0;
           ++ error_count;
           #ifdef errfile
           fprintf(errFile,"error %d\n",log_stat);
           #endif
           -- i;

           // end process if too many errors

           if (error_count == 10)
             break;
         }
         else
         {
           ++ frame_count;
           if (error_count > 0)
             -- error_count;
           if (comp1 == 4)
           {
             log_buffer[frame_byte] = request_id[i];
             log_buffer[frame_byte + 1] = 0xff;
           }
           // no error, so "save" the field in the frame

           if (registered_user == 0 && time_limit != 10000)
           {
             if (log_buffer[frame_byte] == 9)
             {
               if (log_buffer[frame_byte + 1] >= 250)
               {
                 time_limit = 10000;
                 start_time = TntTimer();
               }
             }
           }
//           printf("%d %d/n",i,request_id[i]);
           if (chart_flag == 1)
             (*env)->CallVoidMethod(env, obj, jtid, cable_type, comp1, 0, (int)log_buffer[frame_byte + 1], (int)request_id[i]);
           frame_byte += 2;
         }
       }


       // check for need to flush buffer

       if (frame_byte > 8000)
       {
         fwrite(log_buffer,1,frame_byte,logFile);
         frame_byte = 0;
         final_index = 0;
       }

     }

     #ifdef errfile
     if (error_count == 0)
       display_drb_frame(number_of_parms, &log_buffer[final_index], comp1, errFile);
     #endif

     // calc frame rate

     if (TntTimer() - start_time > 5000)
     {
       frame_rate = (frame_count) / ((TntTimer() - start_time) / 1000);
       (*env)->CallVoidMethod(env, obj, jmid, frame_rate);
       frame_count = 0;
       start_time = TntTimer();
     }

     // get the stop logging indicator

     si = (*env)->GetStaticIntField(env, java_class, fid);
     if (si != last_si)
     {
       (*env)->CallVoidMethod(env, obj, jsid, si);
       if (last_si ==  2)
       {
//         printf("paused\n");
         TntPuts(comp,&request_string[49][0],1);
         pre_request = 0;
       }
       last_si = si;
     }

     // check registered user

     if (registered_user == 0)
     {
//       printf("reg test");
       if (TntTimer() - start_time > time_limit)
         si = 0;
     }

  }
  while (1 == 1 && si != 0 && func != 1 && error_count < 3);
  (*env)->ReleaseIntArrayElements(env, log_parms, body, 0);

  if (func == 2)
  {
   // flush the buffer

   if (final_index > 0)
     fwrite(log_buffer,1,final_index,logFile);

   // close work file

   fclose(logFile);

   // notify java that logging has ceased

  }
//  printf("stopped\n");
  #ifdef errfile
  fclose(errFile);
  #endif
  //remove("logtrace.tmp");
  return (log_stat);
}


jint log_pepper(JNIEnv *env, jobject obj, jint comport, jint func, jint cable_type, jintArray log_parms, jint registered_user)
{
  DWORD start_time;
  DWORD time_limit;
  int i;
  int j;
  int n;
  int number_of_parms;
  int log_field;
  int last_si;
  int frame_count = 0;
  jsize len = (*env)->GetArrayLength(env, log_parms);           // length of array
  jint *body = (*env)->GetIntArrayElements(env, log_parms, 0);  // set pointer to first element

  char request_string[50][20];
  char request_length[50];
  char response_length[50];
  char last_hob[50];
  char last_lob[50];
  char request_priority[50];
  char request_id[50];
  char pre_request;
  char log_buffer[8192];

  int error_count;
  int final_index;
  int frame_byte;
  int frame_rate;
  int log_stat;
  jint si;
  int comp;

  jfieldID fid;
  jmethodID jmid;
  jmethodID jsid;
  jclass java_class;

  FILE *logFile;

  if (comport == 4)
    comp = 0;
  else
    comp = comport;

  // build a 2 dimensional table of request strings and information about
  // then

  number_of_parms = 0;
  request_string[49][0]  = 0xff;

  for (i=0; i<len; i++)
  {
    if (body[i] >= 0)
    {
      last_hob[number_of_parms]           = 0;
      last_lob[number_of_parms]           = 0;
      request_priority[number_of_parms]   = 0;
      request_length[number_of_parms]     = 1;
      response_length[number_of_parms]    = 3;
      request_id[number_of_parms]         = body[i];

      if (body[i] == 0)  // timing
        request_string[number_of_parms][0] = 0xe;
      if (body[i] == 3)  // o2
        request_string[number_of_parms][0] = 0x14;
      if (body[i] == 11 || body[i] == 49)  // iat
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][0] = 0x0f;
      }
      if (body[i] == 1)  // ect
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][0] = 0x05;
      }
      if (body[i] == 12)  // low trim
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][0] = 0x06;
      }
      if (body[i] == 13)  // med trim
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][0] = 0x07;
      }
      if (body[i] == 7)  // tps
        request_string[number_of_parms][0] = 0x11;
      if (body[i] == 18)  // maf
        request_string[number_of_parms][0] = 0x10;
      if (body[i] == 9)  // rpm
        request_string[number_of_parms][0] = 0x0c;
      if (request_string[number_of_parms][0] != 0)
        ++ number_of_parms;
    }
    request_string[49][0] = 0x04;
  }

  // if were are data logging, get java links

  if (func == 2)
  {
    java_class = (*env)->GetObjectClass(env, obj);
    fid = (*env)->GetStaticFieldID(env, java_class, "obd_stop_flag", "I");
    if (fid == 0)
    {
      return(-199);
    }
    jmid = (*env)->GetMethodID(env, java_class, "update_status", "(I)V");
    if (jmid == 0)
    {
      return(-101);
    }
    jsid = (*env)->GetMethodID(env, java_class, "update_pause", "(I)V");
    if (jmid == 0)
    {
      return(-101);
    }
  }

  if (func == 2)
  {
    // open work file

    logFile = fopen("tnt1.tmp", "wb");
    if (logFile < 0)
      return(-100);

    // write number of parms to file

    fwrite(&number_of_parms, sizeof(number_of_parms),1,logFile);

    // write list of parms to file

    for (i=0; i<len; i++)
    {
      if (body[i] >= 0)
        fwrite(&body[i],1,1,logFile);
    }
  }

  // logon to interface

  TntRxClear(comp);
  TntPuts(comp,&request_string[49][0],1);

  // if not logging, return

  if (func != 2)
    return(delayt16(comport, 1, 5));

  frame_byte = 0;
  pre_request = 0;
  error_count = 0;

  start_time = TntTimer();
  time_limit = 120000;
  last_si = 1;

  do
  {
     if (error_count == 0)  // last logged frame was not in error so save it
     {
       final_index = frame_byte;
       pre_request = 0;
     }
     else
       error_count = 0;


     // loop through the send requests, one full pass makes a "frame"
     // an error on any one element, voids the frame
     // three consecutive errors stops the logging process

     if (si != 2)
     {
       for (i=0; i<number_of_parms; i++)
       {

         // get a log field

         if (i == (number_of_parms - 1))
           n = 0;
         else
           n = i + 1;

         log_field = 0;

         // if current field is low pri, chk to see if it's time to run it

         if (frame_count != 0 && request_priority[i] != 0)
         {
           if ((frame_count / request_priority[i]) * request_priority[i] == frame_count)
             log_field = 1;
         }

         if (log_field == 0)
         {

           // log data for field (all fields on frame 0)

           log_stat = getpepperdata(comport, &log_buffer[frame_byte], i, number_of_parms, pre_request, request_length, response_length, request_id, &request_string[i][0], &request_string[n][0], request_length[n]);
//           printf("%d %d\n",log_stat,request_id[i]);
           if (request_priority[i] != 0)
           {

             // save the data for the next frame

             last_hob[i] = log_buffer[frame_byte + 1];
             last_lob[i] = log_buffer[frame_byte + 2];
           }
         }
         else
         {

           // low priority field, so use the same data as last time

           log_stat = 0;
           log_buffer[frame_byte] = request_id[i];
           log_buffer[frame_byte + 1] = last_hob[i];
           log_buffer[frame_byte + 2] = last_lob[i];
           pre_request = 0;
         }

         // test for error

         if (log_stat < 0)
         {
           pre_request = 0;
           ++ error_count;
           -- i;

           // end process if too many errors

           if (error_count == 3)
             break;
         }
         else
         {

           if (comport == 4)
           {
             log_buffer[frame_byte] = request_id[i];
             log_buffer[frame_byte + 1] = 0x00;
             log_buffer[frame_byte + 2] = 0x50;
           }

           // no error, so "save" the field in the frame

           if (registered_user == 0 && time_limit != 10000)
           {
             if (log_buffer[frame_byte] == 9)
             {
               if (log_buffer[frame_byte] + 2 >= 250)
               {
                 time_limit = 10000;
                 start_time = TntTimer();
               }
             }
           }
           frame_byte += 3;
         }
       }

       frame_count += 1; //number_of_parms;

       // check for need to flush buffer

       if (frame_byte > 8000)
       {
         fwrite(log_buffer,1,frame_byte,logFile);
         frame_byte = 0;
         final_index = 0;
       }

     }

     // calc frame rate

     if (error_count == 0)
       display_obd_frame(number_of_parms, &log_buffer[final_index],comport);

//     if (frame_count * number_of_parms > 100 && TntTimer() - start_time > 1000)
//     {
//       frame_rate = (frame_count * number_of_parms) / ((TntTimer() - start_time) / 1000);
       (*env)->CallVoidMethod(env, obj, jmid, frame_rate * number_of_parms);
//     }

     // get the stop logging indicator

     si = (*env)->GetStaticIntField(env, java_class, fid);
     if (si != last_si)
     {
       (*env)->CallVoidMethod(env, obj, jsid, si);
       if (last_si ==  2)
       {
//         printf("paused\n");
         TntPuts(comp,&request_string[49][0],1);
         pre_request = 0;
       }
       last_si = si;
     }

     // check registered user

     if (registered_user == 0)
     {
       if (TntTimer() - start_time > time_limit)
         si = 0;
     }

  }
  while (1 == 1 && si != 0 && func != 1 && error_count < 3);
  (*env)->ReleaseIntArrayElements(env, log_parms, body, 0);

  if (func == 2)
  {
   // flush the buffer

   if (final_index > 0)
     fwrite(log_buffer,1,final_index,logFile);

   // close work file

   fclose(logFile);

   // notify java that logging has ceased

  }
//  printf("stopped\n");
  return (log_stat);
}

jint log_ease(JNIEnv *env, jobject obj, jint comport, jint func, jint cable_type, jintArray log_parms, jint registered_user)
{
  DWORD start_time;
  DWORD time_limit;
  int i;
  int j;
  int n;
  int number_of_parms;
  int log_field;
  int last_si;
  int frame_count = 0;
  jsize len = (*env)->GetArrayLength(env, log_parms);           // length of array
  jint *body = (*env)->GetIntArrayElements(env, log_parms, 0);  // set pointer to first element

  char request_string[50][20];
  char request_length[50];
  char response_length[50];
  char last_hob[50];
  char last_lob[50];
  char request_priority[50];
  char request_id[50];
  char pre_request;
  char log_buffer[8192];

  int error_count;
  int final_index;
  int frame_byte;
  int frame_rate;
  int log_stat;
  jint si;
  int comp;

  jfieldID fid;
  jmethodID jmid;
  jmethodID jsid;
  jclass java_class;

  FILE *logFile;

  if (comport == 4)
    comp = 0;
  else
    comp = comport;

  // build a 2 dimensional table of request strings and information about
  // then

  number_of_parms = 0;
  request_string[49][0]  = 0x06;
  request_string[49][1]  = 0x68;
  request_string[49][2]  = 0x6a;
  request_string[49][3]  = 0xf1;
  request_string[49][4]  = 0x01;
  request_string[49][5]  = 0x0f;
  request_string[49][6]  = 0xd3;

  for (i=0; i<len; i++)
  {
    if (body[i] >= 0)
    {
      last_hob[number_of_parms]           = 0;
      last_lob[number_of_parms]           = 0;
      request_priority[number_of_parms]   = 0;
      request_length[number_of_parms]     = 7;
      response_length[number_of_parms]    = 8;
      request_id[number_of_parms]         = body[i];
      request_string[number_of_parms][0]  = 0x06;
      request_string[number_of_parms][1]  = 0x68;
      request_string[number_of_parms][2]  = 0x6a;
      request_string[number_of_parms][3]  = 0xf1;
      request_string[number_of_parms][4]  = 0x01;

      if (body[i] == 0)  // timing
      {
        request_string[number_of_parms][5] = 0x0e;
        request_string[number_of_parms][6] = 0xd2;
      }
      if (body[i] == 3)  // o2
      {
        request_string[number_of_parms][5] = 0x14;
        request_string[number_of_parms][6] = 0xd8;
        response_length[number_of_parms] = 9;
      }
      if (body[i] == 11 || body[i] == 49)  // iat
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][5] = 0x0f;
        request_string[number_of_parms][6] = 0xd3;
      }
      if (body[i] == 1)  // ect
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][5] = 0x05;
        request_string[number_of_parms][6] = 0xc9;
      }
      if (body[i] == 12)  // s/t
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][5] = 0x06;
        request_string[number_of_parms][6] = 0xca;
      }
      if (body[i] == 13)  // l/t
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][5] = 0x07;
        request_string[number_of_parms][6] = 0xcb;
      }
      if (body[i] == 7)  // tps
      {
        request_string[number_of_parms][5] = 0x11;
        request_string[number_of_parms][6] = 0xd5;
      }
      if (body[i] == 18)  // maf
      {
        request_string[number_of_parms][5] = 0x10;
        request_string[number_of_parms][6] = 0xd4;
        response_length[number_of_parms] = 9;
      }
      if (body[i] == 9)  // rpm
      {
        request_string[number_of_parms][5] = 0x0c;
        request_string[number_of_parms][6] = 0xd0;
        response_length[number_of_parms] = 9;
      }
      if (request_string[number_of_parms][0] != 0)
        ++ number_of_parms;
    }
    request_string[49][5] = 0x11;
    request_string[49][6] = 0xd5;
  }

  // if were are data logging, get java links

  if (func == 2)
  {
    easelogon(comport);

    java_class = (*env)->GetObjectClass(env, obj);
    fid = (*env)->GetStaticFieldID(env, java_class, "obd_stop_flag", "I");
    if (fid == 0)
    {
      return(-199);
    }
    jmid = (*env)->GetMethodID(env, java_class, "update_status", "(I)V");
    if (jmid == 0)
    {
      return(-101);
    }
    jsid = (*env)->GetMethodID(env, java_class, "update_pause", "(I)V");
    if (jmid == 0)
    {
      return(-101);
    }
  }

  if (func == 2)
  {
    // open work file

    logFile = fopen("tnt1.tmp", "wb");
    if (logFile < 0)
      return(-100);

    // write number of parms to file

    fwrite(&number_of_parms, sizeof(number_of_parms),1,logFile);

    // write list of parms to file

    for (i=0; i<len; i++)
    {
      if (body[i] >= 0)
        fwrite(&body[i],1,1,logFile);
    }
  }

  // logon to interface

  TntRxClear(comp);
  TntPuts(comp,&request_string[49][0],7);

  // if not logging, return

  if (func != 2)
    return(delayt16(comport, 5, 5));

  frame_byte = 0;
  pre_request = 0;
  error_count = 0;

  start_time = TntTimer();
  time_limit = 120000;
  last_si = 1;

  do
  {
     if (error_count == 0)  // last logged frame was not in error so save it
     {
       final_index = frame_byte;
       pre_request = 0;
     }
     else
       error_count = 0;


     // loop through the send requests, one full pass makes a "frame"
     // an error on any one element, voids the frame
     // three consecutive errors stops the logging process


//     TntRxClear(0);
     if (si != 2)
     {
       for (i=0; i<number_of_parms; i++)
       {

         // get a log field

         if (i == (number_of_parms - 1))
           n = 0;
         else
           n = i + 1;

         log_field = 0;

         // if current field is low pri, chk to see if it's time to run it

         if (frame_count != 0 && request_priority[i] != 0)
         {
           if ((frame_count / request_priority[i]) * request_priority[i] == frame_count)
             log_field = 1;
         }

         if (log_field == 0)
         {

           // log data for field (all fields on frame 0)

           log_stat = geteasedata(comport, &log_buffer[frame_byte], i, number_of_parms, pre_request, request_length, response_length, request_id, &request_string[i][0], &request_string[n][0], request_length[n]);
//           printf("%d %d %d %d %d\n",log_stat,frame_byte,final_index,i,error_count);
           if (request_priority[i] != 0)
           {

             // save the data for the next frame

             last_hob[i] = log_buffer[frame_byte + 1];
             last_lob[i] = log_buffer[frame_byte + 2];
           }
         }
         else
         {

           // low priority field, so use the same data as last time

           log_stat = 0;
           log_buffer[frame_byte] = request_id[i];
           log_buffer[frame_byte + 1] = last_hob[i];
           log_buffer[frame_byte + 2] = last_lob[i];
           pre_request = 0;
         }

         // test for error

         if (log_stat < 0)
         {
           pre_request = 0;
           ++ error_count;
           -- i;

           // end process if too many errors

           if (error_count == 3)
             break;
         }
         else
         {

           if (comport == 4)
           {
             log_buffer[frame_byte] = request_id[i];
             log_buffer[frame_byte + 1] = 0x00;
             log_buffer[frame_byte + 2] = 0x50;
           }

           // no error, so "save" the field in the frame

           if (registered_user == 0 && time_limit != 10000)
           {
             if (log_buffer[frame_byte] == 9)
             {
               if (log_buffer[frame_byte] + 2 >= 250)
               {
                 time_limit = 10000;
                 start_time = TntTimer();
               }
             }
           }
           frame_byte += 3;
         }
       }

       frame_count += 1; //number_of_parms;

       // calc frame rate

//       if (frame_count * number_of_parms > 100 && TntTimer() - start_time > 1000)
//       {
//         frame_rate = (frame_count * number_of_parms) / ((TntTimer() - start_time) / 1000);
         (*env)->CallVoidMethod(env, obj, jmid, frame_count * number_of_parms);
//       }

       // check for need to flush buffer

       if (frame_byte > 8000)
       {
         fwrite(log_buffer,1,frame_byte,logFile);
         frame_byte = 0;
         final_index = 0;
       }

     }

     // get the stop logging indicator

     if (error_count == 0)
       display_obd_frame(number_of_parms, &log_buffer[final_index],comport);
     si = (*env)->GetStaticIntField(env, java_class, fid);
     if (si != last_si)
     {
       (*env)->CallVoidMethod(env, obj, jsid, si);
       if (last_si ==  2)
       {
         easelogon(comport);
         pre_request = 0;
       }
       last_si = si;
     }

     // check registered user

     if (registered_user == 0)
     {
//       printf("%d\n",time_limit - (TntTimer() - start_time));
       if (TntTimer() - start_time > time_limit)
         si = 0;
     }

  }
  while (1 == 1 && si != 0 && func != 1 && error_count < 3);
  (*env)->ReleaseIntArrayElements(env, log_parms, body, 0);

  if (func == 2)
  {
   // flush the buffer

   if (final_index > 0)
     fwrite(log_buffer,1,final_index,logFile);

   // close work file

   fclose(logFile);

   // notify java that logging has ceased

  }
  return (log_stat);
}

int getpepper99data(jint comport, char log_buffer[], char current_parm, char last_parm, char pre_request, char request_length[], char response_length[], char request_id[], char request_string[], char next_string[], char next_length)
{

  int i_code;
  char buf[20];
  int i1;
  int comp;
  DWORD p1;
  int b;

  if (comport == 4)
    comp = 0;
  else
    comp = comport;

//  TntTxClear(comp);

  // if pre request is zero, request data


//  if (pre_request == 0)
   {

//     TntRxClear(comp);

//     TntPuts(comp,request_string,request_length[current_parm]);

//       xdelay(80);

       TntRxClear(0);
       for (b=0;b<request_length[current_parm];b++)
       {
         TntPutc(0,request_string[b]);
         xdelay(0);
       }
   }

  // delay (wait for data)

  if (delayt16(comport,response_length[current_parm] + 6,5) < 0)
    return(-1);
//  else
//    pre_request = 1;

  // request NEXT element and set pre-request to 1

  if (comport == 4)
    return(0);

  do
  {

    // scan for header

    i_code = TntGetc(comp);

    if (i_code == 0x48)  // start message frame
    {

//      TntPuts(comp,next_string,next_length);

      //  block header found, so stop the search

      TntGets(comp,(LPSTR)buf,4); // skip the balance of the header

      // get the data bytes

      log_buffer[0] = request_id[current_parm];
      if (request_id[current_parm] != 9 &&
//          request_id[current_parm] != 3 &&
          request_id[current_parm] != 18)
      {
        log_buffer[1] = 0;
        log_buffer[2] = TntGetc(comp);
      }
      else
      {
        log_buffer[1] = TntGetc(comp);
        log_buffer[2] = TntGetc(comp);
      }
      TntGetc(comp);
      xdelay(70);
      return(0);

    }
  }
  while (TntRxQue(comp) > 0);

  // we get here ONLY if we failed to find a data block

  return (-2);
}

jint log_pepper99(JNIEnv *env, jobject obj, jint comport, jint func, jint cable_type, jintArray log_parms, jint registered_user)
{
  DWORD start_time;
  DWORD time_limit;
  int i;
  int j;
  int n;
  int number_of_parms;
  int log_field;
  int last_si;
  int frame_count = 0;
  jsize len = (*env)->GetArrayLength(env, log_parms);           // length of array
  jint *body = (*env)->GetIntArrayElements(env, log_parms, 0);  // set pointer to first element

  char request_string[50][20];
  char request_length[50];
  char response_length[50];
  char last_hob[50];
  char last_lob[50];
  char request_priority[50];
  char request_id[50];
  char pre_request;
  char log_buffer[8192];

  int error_count;
  int final_index;
  int frame_byte;
  int frame_rate;
  int log_stat;
  jint si;
  int comp;

  jfieldID fid;
  jmethodID jmid;
  jmethodID jsid;
  jclass java_class;

  FILE *logFile;

  if (comport == 4)
    comp = 0;
  else
    comp = comport;

  // build a 2 dimensional table of request strings and information about
  // then

  number_of_parms = 0;
  request_string[49][0]  = 0x68;
  request_string[49][1]  = 0x6a;
  request_string[49][2]  = 0xf1;
  request_string[49][3]  = 0x01;
  request_string[49][4]  = 0x0f;
  request_string[49][5]  = 0xd3;

  for (i=0; i<len; i++)
  {
    if (body[i] >= 0)
    {
      last_hob[number_of_parms]           = 0;
      last_lob[number_of_parms]           = 0;
      request_priority[number_of_parms]   = 0;
      request_length[number_of_parms]     = 6;
      response_length[number_of_parms]    = 7;
      request_id[number_of_parms]         = body[i];
      request_string[number_of_parms][0]  = 0x68;
      request_string[number_of_parms][1]  = 0x6a;
      request_string[number_of_parms][2]  = 0xf1;
      request_string[number_of_parms][3]  = 0x01;

      if (body[i] == 0)  // timing
      {
        request_string[number_of_parms][4] = 0x0e;
        request_string[number_of_parms][5] = 0xd2;
      }
      if (body[i] == 3)  // o2
      {
        request_string[number_of_parms][4] = 0x14;
        request_string[number_of_parms][5] = 0xd8;
//      response_length[number_of_parms] = 8;
      }
      if (body[i] == 11 || body[i] == 49)  // iat
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][4] = 0x0f;
        request_string[number_of_parms][5] = 0xd3;
      }
      if (body[i] == 1)  // ect
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][4] = 0x05;
        request_string[number_of_parms][5] = 0xc9;
      }
      if (body[i] == 12)  // s/t
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][4] = 0x06;
        request_string[number_of_parms][5] = 0xca;
      }
      if (body[i] == 13)  // l/t
      {
//        request_priority[number_of_parms]  = 10;
        request_string[number_of_parms][4] = 0x07;
        request_string[number_of_parms][5] = 0xcb;
      }
      if (body[i] == 7)  // tps
      {
        request_string[number_of_parms][4] = 0x11;
        request_string[number_of_parms][5] = 0xd5;
      }
      if (body[i] == 18)  // maf
      {
        request_string[number_of_parms][4] = 0x10;
        request_string[number_of_parms][5] = 0xd4;
        response_length[number_of_parms] = 8;
      }
      if (body[i] == 9)  // rpm
      {
        request_string[number_of_parms][4] = 0x0c;
        request_string[number_of_parms][5] = 0xd0;
        response_length[number_of_parms] = 8;
      }
      if (request_string[number_of_parms][0] != 0)
        ++ number_of_parms;
    }
    request_string[49][4] = 0x06;
    request_string[49][5] = 0xca;
  }

  // if were are data logging, get java links

  if (func == 2)
  {
    xdelay(200);
    pepperlogon(comport);

    java_class = (*env)->GetObjectClass(env, obj);
    fid = (*env)->GetStaticFieldID(env, java_class, "obd_stop_flag", "I");
    if (fid == 0)
    {
      return(-199);
    }
    jmid = (*env)->GetMethodID(env, java_class, "update_status", "(I)V");
    if (jmid == 0)
    {
      return(-101);
    }
    jsid = (*env)->GetMethodID(env, java_class, "update_pause", "(I)V");
    if (jmid == 0)
    {
      return(-101);
    }
  }

  if (func == 2)
  {
    // open work file

    logFile = fopen("tnt1.tmp", "wb");
    if (logFile < 0)
      return(-100);

    // write number of parms to file

    fwrite(&number_of_parms, sizeof(number_of_parms),1,logFile);

    // write list of parms to file

    for (i=0; i<len; i++)
    {
      if (body[i] >= 0)
        fwrite(&body[i],1,1,logFile);
    }
  }

  // logon to interface

  TntRxClear(comp);
//  if (func != 2)
//    TntPuts(comp,&request_string[49][0],6);

  // if not logging, return

  if (func != 2)
    return(0);
//    return(delayt16(comport, 13, 5));
  else
    xdelay(2000);

  frame_byte = 0;
  pre_request = 0;
  error_count = 0;

  start_time = TntTimer();
  time_limit = 120000;
  last_si = 1;

  do
  {
     if (error_count == 0)  // last logged frame was not in error so save it
     {
       final_index = frame_byte;
       pre_request = 0;
     }
     else
       error_count = 0;


     // loop through the send requests, one full pass makes a "frame"
     // an error on any one element, voids the frame
     // three consecutive errors stops the logging process


//     TntRxClear(0);
     if (si != 2)
     {
       for (i=0; i<number_of_parms; i++)
       {

         // get a log field

         if (i == (number_of_parms - 1))
           n = 0;
         else
           n = i + 1;

         log_field = 0;

         // if current field is low pri, chk to see if it's time to run it

         if (frame_count != 0 && request_priority[i] != 0)
         {
           if ((frame_count / request_priority[i]) * request_priority[i] == frame_count)
             log_field = 1;
         }

         if (log_field == 0)
         {

           // log data for field (all fields on frame 0)

           log_stat = getpepper99data(comport, &log_buffer[frame_byte], i, number_of_parms, pre_request, request_length, response_length, request_id, &request_string[i][0], &request_string[n][0], request_length[n]);
//           printf("%d %d %d %d %d\n",log_stat,frame_byte,final_index,TntTimer(),error_count);
           if (request_priority[i] != 0)
           {

             // save the data for the next frame

             last_hob[i] = log_buffer[frame_byte + 1];
             last_lob[i] = log_buffer[frame_byte + 2];
           }
         }
         else
         {

           // low priority field, so use the same data as last time

           log_stat = 0;
           log_buffer[frame_byte] = request_id[i];
           log_buffer[frame_byte + 1] = last_hob[i];
           log_buffer[frame_byte + 2] = last_lob[i];
           pre_request = 0;
         }

         // test for error

         if (log_stat < 0)
         {
           pre_request = 0;
           ++ error_count;
           -- i;

           // end process if too many errors

           if (error_count == 3)
             break;
         }
         else
         {

           if (comport == 4)
           {
             log_buffer[frame_byte] = request_id[i];
             log_buffer[frame_byte + 1] = 0x00;
             log_buffer[frame_byte + 2] = 0x50;
           }

           // no error, so "save" the field in the frame

           if (registered_user == 0 && time_limit != 10000)
           {
             if (log_buffer[frame_byte] == 9)
             {
               if (log_buffer[frame_byte] + 2 >= 250)
               {
                 time_limit = 10000;
                 start_time = TntTimer();
               }
             }
           }
           frame_byte += 3;
         }
       }

       frame_count += 1; //number_of_parms;

       // calc frame rate

//       if (frame_count * number_of_parms > 100 && TntTimer() - start_time > 1000)
//       {
//         frame_rate = (frame_count * number_of_parms) / ((TntTimer() - start_time) / 1000);
         (*env)->CallVoidMethod(env, obj, jmid, frame_count * number_of_parms);
//       }

       // check for need to flush buffer

       if (frame_byte > 8000)
       {
         fwrite(log_buffer,1,frame_byte,logFile);
         frame_byte = 0;
         final_index = 0;
       }

     }

     // get the stop logging indicator

     if (error_count == 0)
       display_obd_frame(number_of_parms, &log_buffer[final_index],comport);
     si = (*env)->GetStaticIntField(env, java_class, fid);
     if (si != last_si)
     {
       (*env)->CallVoidMethod(env, obj, jsid, si);
       if (last_si ==  2)
       {
         easelogon(comport);
         pre_request = 0;
       }
       last_si = si;
     }

     // check registered user

     if (registered_user == 0)
     {
//       printf("%d\n",time_limit - (TntTimer() - start_time));
       if (TntTimer() - start_time > time_limit)
         si = 0;
     }

  }
  while (1 == 1 && si != 0 && func != 1 && error_count < 3);
  (*env)->ReleaseIntArrayElements(env, log_parms, body, 0);

  if (func == 2)
  {
   // flush the buffer

   if (final_index > 0)
     fwrite(log_buffer,1,final_index,logFile);

   // close work file

   fclose(logFile);

   // notify java that logging has ceased

  }
  return (log_stat);
}

JNIEXPORT jint JNICALL
Java_serialio_logdata(JNIEnv *env, jobject obj, jint comport, jint func, jint cable, jintArray arr, jint reg, jint chart_flag)
{
//  printf("logdata %d\n",cable);
  if (cable == 4)
   return(log_drb(env, obj, comport, func, cable, arr, reg, chart_flag));
  if (cable == 0)
   return(log_pepper(env, obj, comport, func, cable, arr, reg));
  if (cable == 3)
   return(log_ease(env, obj, comport, func, cable, arr, reg));
  if (cable == 1)
   return(log_t16(env, obj, comport, func, cable, arr, reg, chart_flag));
  if (cable == 2)
   return(log_pepper99 (env, obj, comport, func, cable, arr, reg));
}

void display_obd_frame(int element_count, unsigned char element_array[], int func)
{
  int l1;
  int b1 = 0;
  char field_id;
  float f1;
  int i1;

//  if (element_count < 1)
    return;

  for (l1=0;l1<element_count;l1++)
  {
    field_id = element_array[b1];
    if (field_id == 9)
    {
      if (func == 4)
      {
        element_array[b1 + 1] = 0x3e;
        element_array[b1 + 2] = 0x80;
      }
      i1 = ((element_array[b1 + 1] * 256) + element_array[b1 + 2]) / 4;
      printf("Rpm: %d ",i1);
    }
    else
    if (field_id == 3)
    {
      if (func == 4)
      {
        element_array[b1 + 2] = 150;
      }
      f1 = element_array[b1 + 2] * .005F;
      printf("O2v: %.2f ",f1);
    }
    else
    if (field_id == 0)
    {
      if (func == 4)
      {
        element_array[b1 + 2] = 180;
      }
      f1 = (element_array[b1 + 2] - 128.0F ) * .5F;
      printf("Timing: %.0f ",f1);
    }
    else
    if (field_id == 12)
    {
      if (func == 4)
      {
        element_array[b1 + 2] = 150;
      }
      f1 = (element_array[b1 + 2] / 128.0F ) * 100.0F - 100.0F;
      printf("S trim: %.0f ",f1);
    }
    else
    if (field_id == 13)
    {
      if (func == 4)
      {
        element_array[b1 + 2] = 180;
      }
      f1 = (element_array[b1 + 2] / 128.0F ) * 100.0F - 100.0F;
      printf("L trim: %.0f ",f1);
    }
    else
    if (field_id == 7)
    {
      f1 = (element_array[b1 + 2] / 255.0F ) * 100.0F;
      printf("Tps: %.0f ",f1);
    }
    else
    if (field_id == 18)
    {
      f1 = ((element_array[b1 + 1] * 256) + element_array[b1 + 2]) * .132275F;
      printf("Air Flow: %.2f ",f1);
    }
    else
    if (field_id == 1)
    {
      if (func == 4)
      {
        element_array[b1 + 2] = 150;
      }
      f1 = (element_array[b1 + 2] - 40.0F) * 1.8F + 32.0F;
      printf("Ect: %.0f ",f1);
    }
    else
    if (field_id == 11)
    {
      if (func == 4)
      {
        element_array[b1 + 2] = 190;
      }
      f1 = (element_array[b1 + 2] - 40.0F) * 1.8F + 32.0F;
      printf("Iat: %.0f ",f1);
    }
    else
      printf("Field %d ",field_id);
    b1 += 3;
  }
  printf("\n");
}

void display_drb_frame(int element_count, unsigned char element_array[], int func, FILE *errFile)
{
  int l1;
  int b1 = 0;
  char field_id;
  float f1;
  int i1;

  if (element_count < 1)
    return;

  for (l1=0;l1<element_count;l1++)
  {
    field_id = element_array[b1];
    if (field_id == 9)
    {
      if (func == 4)
      {
        element_array[b1 + 1] = 127;
      }
      f1 = element_array[b1 + 1] * 31.37F;
      fprintf(errFile,"Rpm: %.0f ",f1);
    }
    else
    if (field_id == 3)
    {
      if (func == 4)
      {
        element_array[b1 + 1] = 150;
      }
      f1 = element_array[b1 + 1] * .02F;
      fprintf(errFile,"O2v: %.2f ",f1);
    }
    else
    if (field_id == 0)
    {
      if (func == 4)
      {
        element_array[b1 + 1] = 180;
      }
      f1 = element_array[b1 + 1] - 10.0F;
      fprintf(errFile,"Timing: %.0f %d",f1,element_array[b1 + 1]);
    }
    else
    if (field_id == 12)
    {
      if (func == 4)
      {
        element_array[b1 + 1] = 150;
      }
      f1 = (element_array[b1 + 1] * 1.27F);
      fprintf(errFile,"S trim: %.0f ",f1);
    }
    else
    if (field_id == 13)
    {
      if (func == 4)
      {
        element_array[b1 + 1] = 180;
      }
      f1 = (element_array[b1 + 1] * 1.27F);
      fprintf(errFile,"L trim: %.0f ",f1);
    }
    else
    if (field_id == 48)
    {
      if (func == 4)
      {
        element_array[b1 + 1] = 180;
      }
      f1 = (element_array[b1 + 1] * 1.27F);
      fprintf(errFile,"M trim: %.0f ",f1);
    }
    else
    if (field_id == 7)
    {
      f1 = (element_array[b1 + 1] / 2.5F );
      fprintf(errFile,"Tps: %.0f ",f1);
    }
    else
    if (field_id == 18)
    {
      f1 = element_array[b1 + 1];
      fprintf(errFile,"Air Flow: %.2f ",f1);
    }
    else
    if (field_id == 1)
    {
      if (func == 4)
      {
        element_array[b1 + 1] = 150;
      }
      f1 = 198 - element_array[b1 + 1];
      fprintf(errFile,"Ect: %.0f ",f1);
    }
    else
    if (field_id == 11)
    {
      if (func == 4)
      {
        element_array[b1 + 1] = 190;
      }
      f1 = 211 - element_array[b1 + 1];
      fprintf(errFile,"Iat: %.0f ",f1);
    }
    else
    if (field_id == 10)
    {
      if (func == 4)
      {
        element_array[b1 + 1] = 128;

      }
      f1 = element_array[b1 + 1] * .256F;
      fprintf(errFile,"Pw: %.2f ",f1);
    }
    else
    if (field_id == 17)
    {
      if (func == 4)
      {
        element_array[b1 + 1] = 40;
      }
      f1 = element_array[b1 + 1];
      fprintf(errFile,"Knck: %.2f ",f1);
    }
    else
      fprintf(errFile,"Field %d ",field_id);
    b1 += 2;
  }
  fprintf(errFile,"\n");
}

int
getpns(JNIEnv *env, jobject obj,int a)
{
  unsigned char data_buf[10];
  int parm_count;
  int parm_size;
  unsigned char parm_list[50];
  int v1;

  jmethodID jmid;
  jclass java_class;
  FILE *logFile;

  java_class = (*env)->GetObjectClass(env, obj);
  jmid = (*env)->GetMethodID(env, java_class, "update_pnstein", "(I)V");
  if (jmid == 0)
  {
    return(-101);
  }

  // open file

  logFile = fopen("tnt1.tmp", "wb");
  if (logFile < 0)
    return(-100);

  TntBaud(a, 19200);
  TntParms(a,0,0,8);
  TntFlow(a,'N');
  v1 = TntReset(a,1024,1024);
  if (v1 != 0)
    return(v1);

  // wait for xff

//   printf("waiting on 0xff\n");

  (*env)->CallVoidMethod(env, obj, jmid, 1);

  if (delayt16(a,1,30) == 0)
  {
    // is it a 0xFF?

    TntGets(a,(LPSTR)data_buf,1);

    (*env)->CallVoidMethod(env, obj, jmid, 1);

//    printf("Got %d\n",data_buf[0]);

    if (data_buf[0] == 0xFF)
    {
      // send 0x00

      data_buf[0] = 0x00;

      TntRxClear(a);
      TntPuts(a,data_buf,1);
      (*env)->CallVoidMethod(env, obj, jmid, 1);

     // printf("Sent 0x00, waiting on parm count\n");

      // wait for parm count

      if (delayt16(a,1,5) == 0)
      {

       // get the parm count

       TntGets(a,(LPSTR)data_buf,1);
       fwrite(data_buf,1,1,logFile);
       (*env)->CallVoidMethod(env, obj, jmid, 1);

       parm_size = (int)data_buf[0];

       TntGets(a,(LPSTR)data_buf,1);
       fwrite(data_buf,1,1,logFile);

       parm_count = (int)data_buf[0];

      // printf("got a parm count of %d\n",parm_count);

       for(v1=0;v1<parm_count;v1++)
       {

         // get the parm list

        // printf("getting parms\n");

         if (delayt16(a,1,5) == 0)
         {
           TntGets(a,&parm_list[v1],1);

          // printf(" %d ",parm_list[v1]);
           fwrite(&parm_list[v1],1,1,logFile);
           TntPuts(a,&parm_list[v1],1);
           (*env)->CallVoidMethod(env, obj, jmid, 1);
         }
       }

//       printf("%d\n",parm_size + 1);
       // read the data blocks

//       getch();
       v1 = 0;
       data_buf[0] = 0xff;
       do
       {
//         printf("*\n");
         if (delayt16(a,parm_size + 1,5) == 0)
         {
           TntGets(a,data_buf,parm_size + 1);
          // printf("%d %d %d\n",data_buf[0],data_buf[1],data_buf[2]);
           if (data_buf[0] != 0xff)
           {
             // write data file
             (*env)->CallVoidMethod(env, obj, jmid, parm_size);
             for (v1=0;v1<parm_size+1;v1++)
               fwrite(&data_buf[v1],1,1,logFile);
           }
           TntTxClear(a);
           TntPuts(a,data_buf,1);
         }
         else
         {
           fclose(logFile);
           break;
         }
       }
       while (1 == 1);
//       printf("%d\n",v1);

      }
    }
  }

  // close file

  fclose(logFile);

  TntDone(a);

  return(0);
}

JNIEXPORT jint JNICALL
Java_tunerlib_getPNSTEIN(JNIEnv *env, jobject obj, jint a)
{
  return getpns(env, obj, a);
}

int getdtc(JNIEnv *env, jobject obj,int a)
{

  unsigned char request_string[20];
  unsigned char data_string[20];
  unsigned char message_string[100];
  unsigned char dtc_string[200];
  int read_length;
  unsigned char byte1;
  unsigned char dtc_list[10][10];
  int m;
  int tint;
  int m1;
  int dtc_count;

  FILE *logFile;
  FILE *dtcFile;

  dtcFile = fopen("dtc.txt","rb");

  logFile = fopen("dtc.tmp", "wb");
  if (logFile < 0)
    return(-100);

  // build the request string;

  request_string[0]  = 0x25;
  request_string[1]  = 0x01;
  request_string[2]  = 0x05;
  request_string[3]  = 0x68;
  request_string[4]  = 0x6a;
  request_string[5]  = 0xf1;
  request_string[6]  = 0x01;
  request_string[7]  = 0x01;
  request_string[8]  = 0x00;
  request_string[9]  = 0x00;
  request_string[10] = 0x00;
  request_string[11] = 0x00;
  request_string[12] = 0x00;
  request_string[13] = 0x00;
  request_string[14] = 0xcb;

  sprintf(message_string,"Unable to connect to the vehicle");

  // delay 5 seconds

  xdelay(5000);

  TntRxClear(a);
  TntPuts(a, request_string,15);

  if (delayt16(a, 14, 5) == 0 || a == 4)
  {

    // request dtc list

    if (a == 4)
    {
        data_string[0] = 0x48;
        data_string[7] = 0x83;
        tint = 8;
    }
    else
      tint = TntGets(a, (LPSTR)data_string,14);

    for (m=0;m<tint;m++)
      if (data_string[m] == 0x48)
      {
         byte1 = data_string[7];
         break;
      }
                                 
    if (m != tint)
    {

      if (byte1 > 128)
        dtc_count = byte1 - 128;
      else
        dtc_count = byte1;
      sprintf(message_string,"%d error codes found, C/E light is ",dtc_count);
      if (byte1 > 128)
        strcat(message_string,"on");
      else
        strcat(message_string,"off");

    }

    byte1 = dtc_count;
    fwrite(&byte1,1,1,logFile);
    if (dtc_count > 0)
      fprintf(logFile,"%s\n",message_string);

    if (dtc_count == 0)
    {
      fprintf(logFile,"No error codes found\n");
    }
    else
    {
       request_string[0]  = 0x25;
       request_string[1]  = 0x10;
       request_string[2]  = 0x05;
       request_string[3]  = 0x68;
       request_string[4]  = 0x6a;
       request_string[5]  = 0xf1;
       request_string[6]  = 0x03;
       request_string[7]  = 0x00;
       request_string[8]  = 0x00;
       request_string[9]  = 0x00;
       request_string[10] = 0x00;
       request_string[11] = 0x00;
       request_string[12] = 0x00;
       request_string[13] = 0x00;
       request_string[14] = 0xdb;

       TntRxClear(a);
       TntPuts(a, request_string,15);

       if (delayt16(a, 14, 5) == 0 || a == 4)
       {

         do
         {
           // get first block of dtc's

           if (a != 4)
             TntGets(a,(LPSTR)data_string,14);
           else
           {
             tint = 12;
             data_string[2]  = 0x48;
             data_string[6]  = 0x01;
             data_string[7]  = 0x43;
             data_string[8]  = 0x01;
             data_string[9]  = 0x96;
             data_string[10] = 0xff;
             data_string[11] = 0x34;

           }

           for (m=0;m<tint;m++)
             if (data_string[m] == 0x48)
             {
                byte1 = data_string[m + 4];
                break;
             }

           m1 = 0;

           if (m != tint)
           {
             m = 6;
             if (dtcFile >= 0)
               fseek(dtcFile,0L,SEEK_SET);

             do
             {
               if ((data_string[m] * 256) + data_string[m+1] == 0)
                break;

               // format the dtc code

               if (data_string[m] >= 192)
               {
                 dtc_list[m1][0] = 'U';
                 data_string[m] -= 192;
               }
               else
                 if (data_string[m] >= 128)
                 {
                   dtc_list[m1][0] = 'B';
                   data_string[m] -= 128;
                 }
                 else
                   if (data_string[m] >= 64)
                   {
                     dtc_list[m1][0] = 'C';
                     data_string[m] -= 64;
                   }
                   else
                     {
                        dtc_list[m1][0] = 'P';
                     }
                sprintf(&dtc_list[m1][1],"%.4x",(data_string[m] * 256) + data_string[m + 1]);

                // get the dtc description

                if (dtcFile < 0)
                  fprintf(logFile,"%s\n",&dtc_list[m1][0]);
                else
                do
                {
                  read_length = 200;
                  if (fgets(dtc_string,read_length,dtcFile) == NULL)
                  {
                    fprintf(logFile,"%s",&dtc_list[m1][0]);
                    sprintf(dtc_string," Unknown DTC code, contact the author");
                    fprintf(logFile,"%s\n",dtc_string);
                    break;
                  }
                  if (strlen(dtc_string) > 5)
                    if (dtc_string[0] == dtc_list[m1][0] &&
                        dtc_string[1] == dtc_list[m1][1] &&
                        dtc_string[2] == dtc_list[m1][2] &&
                        dtc_string[3] == dtc_list[m1][3] &&
                        dtc_string[4] == dtc_list[m1][4])
                    {
                      fprintf(logFile,"%s",&dtc_string[0]);
                      break;
                    }
                }
                while (1 == 1);

                m += 2;
                m1 ++;
             }
             while (m < 11);

           }

           if (TntRxQue(a) < 14)
             break;
         }
         while (tint > 0);
       }


    }
  }

  fclose(logFile);
  if (dtcFile >= 0)
    fclose(dtcFile);
  return 0;
}

int cleardtc(JNIEnv *env, jobject obj,int a)
{

  unsigned char request_string[20];
  unsigned char data_string[20];
  unsigned char message_string[100];
  unsigned char dtc_string[200];
  int read_length;
  unsigned char byte1;
  unsigned char dtc_list[10][10];
  int m;
  int tint;
  int m1;
  int dtc_count;

  FILE *logFile;
  FILE *dtcFile;

  // build the request string;

  request_string[0]  = 0x25;
  request_string[1]  = 0x01;
  request_string[2]  = 0x05;
  request_string[3]  = 0x68;
  request_string[4]  = 0x6a;
  request_string[5]  = 0xf1;
  request_string[6]  = 0x04;
  request_string[7]  = 0x00;
  request_string[8]  = 0x00;
  request_string[9]  = 0x00;
  request_string[10] = 0x00;
  request_string[11] = 0x00;
  request_string[12] = 0x00;
  request_string[13] = 0x00;
  request_string[14] = 0xcd;

  sprintf(message_string,"Unable to connect to the vehicle");

  // delay 5 seconds

  xdelay(5000);

  TntRxClear(a);
  TntPuts(a, request_string,15);

  if (delayt16(a, 14, 5) == 0 || a == 4)
  {

    // request dtc list

    if (a == 4)
    {
        data_string[0] = 0x44;
        tint = 8;
    }
    else
      tint = TntGets(a, (LPSTR)data_string,14);

    for (m=0;m<tint;m++)
      if (data_string[m] == 0x44)
        return(0);
  }
  return(-1);
}

int get1gdtc(JNIEnv *env, jobject obj, jint p)
{
  int port;
  int x;
  int dtc_bits;
  unsigned char dtc_count[1];
  unsigned char out_byte[1];
  unsigned char in_bytes[2];
  char dtc_messages[16][80];
  FILE *logFile;
  fpos_t fp;

  dtc_count[0] = 0;

  if (p == 4)
    port = 0;
  else
    port = p;

  // clear rx buffer

  if (p == 4)
    dtc_bits = 0xffff;
  else
  {

  TntRxClear(port);

  // enable dtc mode

  for (x=0;x<3;x++)
  {
   out_byte[0] = 0xfe;
   TntRxClear(port);
   TntPuts(port,out_byte,1);
   do
   {
     if (delayt16(p,1,2) == 0)
     {
       out_byte[0] = TntGetc(port);
       if (out_byte[0] = 0xfe)
       {
         x = 3;
         break;
       }
     }
     else
       break;  
   } while (1 == 1);
  }

  if (out_byte[0] != 0xfe)
    return -151;

    // get dtc byte 1

  for (x=0;x<3;x++)
  {
   out_byte[0] = 0x3b;
   TntRxClear(port);
   TntPuts(port,out_byte,1);
   do
   {
     if (delayt16(p,1,2) == 0)
     {
       out_byte[0] = TntGetc(port);
       if (out_byte[0] = 0x3b)
       {
         x = 3;
         break;
       }
     }
     else
       break;  
   } while (1 == 1);
  }

  if (out_byte[0] != 0x3b)
    return -152;

  if (delayt16(p,1,1) != 0)
    return -153;

  in_bytes[1] = TntGetc(port);

  dtc_bits = in_bytes[1] * 256;


    // get dtc byte 2

  for (x=0;x<3;x++)
  {
   out_byte[0] = 0x3c;
   TntRxClear(port);
   TntPuts(port,out_byte,1);
   do
   {
     if (delayt16(p,1,2) == 0)
     {
       out_byte[0] = TntGetc(port);
       if (out_byte[0] = 0x3c)
       {
         x = 3;
         break;
       }
     }
     else
       break;  
   } while (1 == 1);
  }

  if (out_byte[0] != 0x3c)
    return -162;

  if (delayt16(p,1,1) != 0)
    return -163;

  in_bytes[1] = TntGetc(port);

  dtc_bits += in_bytes[1];

  for (x=0;x<3;x++)
  {
   out_byte[0] = 0xfd;
   TntRxClear(port);
   TntPuts(port,out_byte,1);
   do
   {
     if (delayt16(p,1,2) == 0)
     {
       out_byte[0] = TntGetc(port);
       if (out_byte[0] = 0xfd)
       {
         x = 3;
         break;
       }
     }
     else
       break;  
   } while (1 == 1);
  }

  if (out_byte[0] != 0xfd)
    return -171;
  }
  // build response file

  logFile = fopen("dtc.tmp", "wb");
  if (logFile < 0)
    return(-100);

  // 32768

  if (dtc_bits >= 32768)
  {
    dtc_bits -= 32768;
    sprintf(&dtc_messages[dtc_count[0]][0],"44 - Ignition system");
    ++ dtc_count[0];
  }

  // 16384
  if (dtc_bits >= 16384)
  {
    dtc_bits -= 16384;
    sprintf(&dtc_messages[dtc_count[0]][0],"43 - EGR system");
    ++ dtc_count[0];
  }

  // 8192
  if (dtc_bits >= 8192)
  {
    dtc_bits -= 8192;
    sprintf(&dtc_messages[dtc_count[0]][0],"42 - Fuel pump");
    ++ dtc_count[0];
  }

  // 4096
  if (dtc_bits >= 4096)
  {
    dtc_bits -= 4096;
    sprintf(&dtc_messages[dtc_count[0]][0],"41 - Fuel injector");
    ++ dtc_count[0];
  }

  // 2048
  if (dtc_bits >= 2048)
  {
    dtc_bits -= 2048;
    sprintf(&dtc_messages[dtc_count[0]][0],"36 - Ignition timing");
    ++ dtc_count[0];
  }

  // 1024
  if (dtc_bits >= 1024)
  {
    dtc_bits -= 1024;
    sprintf(&dtc_messages[dtc_count[0]][0],"31 - Knock sensor");
    ++ dtc_count[0];
  }

  // 512
  if (dtc_bits >= 512)
  {
    dtc_bits -= 512;
    sprintf(&dtc_messages[dtc_count[0]][0],"25 - Barometric sensor");
    ++ dtc_count[0];
  }

  // 256
  if (dtc_bits >= 256)
  {
    dtc_bits -= 256;
    sprintf(&dtc_messages[dtc_count[0]][0],"24 - Speed sensor");
    ++ dtc_count[0];
  }

  // 128
  if (dtc_bits >= 128)
  {
    dtc_bits -= 128;
    sprintf(&dtc_messages[dtc_count[0]][0],"23 - Cam sensor");
    ++ dtc_count[0];
  }

  // 64
  if (dtc_bits >= 64)
  {
    dtc_bits -= 64;
    sprintf(&dtc_messages[dtc_count[0]][0],"22 - Crank sensor");
    ++ dtc_count[0];
  }

  // 32
  if (dtc_bits >= 32)
  {
    dtc_bits -= 32;
    sprintf(&dtc_messages[dtc_count[0]][0],"21 - Engine coolant");
    ++ dtc_count[0];
  }

  // 16
  if (dtc_bits >= 16)
  {
    dtc_bits -= 16;
    sprintf(&dtc_messages[dtc_count[0]][0],"15 - ISC");
    ++ dtc_count[0];
  }

  // 8
  if (dtc_bits >= 8)
  {
    dtc_bits -= 8;
    sprintf(&dtc_messages[dtc_count[0]][0],"14 - TPS");
    ++ dtc_count[0];
  }

  // 4
  if (dtc_bits >= 4)
  {
    dtc_bits -= 4;
    sprintf(&dtc_messages[dtc_count[0]][0],"13 - Intake air temp");
    ++ dtc_count[0];
  }

  // 2
  if (dtc_bits >= 2)
  {
    dtc_bits -= 2;
    sprintf(&dtc_messages[dtc_count[0]][0],"12 - Air flow sensor");
    ++ dtc_count[0];
  }

  // 1

  if (dtc_bits >= 1)
  {
    dtc_bits -= 1;
    sprintf(&dtc_messages[dtc_count[0]][0],"11 - O2 sensor");
    ++ dtc_count[0];
  }

  fwrite(&dtc_count[0],1,1,logFile);

  if (dtc_count[0] == 0)
    fprintf(logFile,"No trouble codes found\n");
  else
  {
    fprintf(logFile,"%d trouble codes found\n",dtc_count[0]);
    for (x=0;x<dtc_count[0];x++)
    {
      fprintf(logFile,"%s\n",&dtc_messages[x][0]);
    }
  }

  fclose(logFile);
  return 0;
}

JNIEXPORT jint JNICALL
Java_serialio_getDTC(JNIEnv *env, jobject obj, jint a, jint c)
{
  if (c == 4) // 1g dsm
    return get1gdtc(env, obj, a);
  else
    return getdtc(env, obj, a);
}

JNIEXPORT jint JNICALL
Java_serialio_clearDTC(JNIEnv *env, jobject obj, jint a)
{
  return cleardtc(env, obj, a);
}

JNIEXPORT jint JNICALL
Java_serialio_getver(JNIEnv *env, jobject obj)
{
  return(6);
}

JNIEXPORT jint JNICALL
Java_tunerlib_getver(JNIEnv *env, jobject obj)
{
  return(6);
}

JNIEXPORT jint JNICALL
Java_serialio_dodiag(JNIEnv *env, jobject obj, jint port, jint diag)
{
  char buf[2];
  int p;
  int x;
  char command_on[11] = {0xf6,0xf4,0xf1,0xf3,0xf5,0xfc,0xfb,0xfa,0xf9,0xf8,0xf7};

  if (port == 4)
    p = 0;
  else
    p = port;

//  TntParms(p, 0, 0, 8);
//  TntFlow(p, 'N');
//  TntBaud(p,1920);

  switch(diag)
  {
    case 99:
      buf[0] = 0xff;
      break;
    default:
      buf[0] = command_on[diag];
      break;
  }

  if (p == 4)
    return 0;

  for (x=0;x<3;x++)
  {
   TntRxClear(port);
   TntPuts(port,&buf[0],1);
   do
   {
     if (delayt16(p,1,2) == 0)
     {
       buf[1] = TntGetc(port);
       if (buf[0] = buf[1])
       {
         x = 3;
         break;
       }
     }
     else
       break;  
   } while (1 == 1);
  }

  if (buf[0] != buf[1])
    return -151;
  return(0);
}

void main()
{
  TntBaud(0, 19200);
  TntParms(0,0,0,8);
  TntFlow(0,'N');
  TntReset(0,1024,1024);


//  getdtc(NULL,NULL,4);

  TntDone(0);
}
