import java.awt.*;



class CubeDrawer extends Object {


    private float []  contourx;
    private float []  contoury;
    private float     centerx;
    private float     centery;


    private final static int contourtable [] [] = {
        {4, 5, 1, 3, 2, 6},
        {5, 7, 3, 2, 0, 4},
        {6, 4, 0, 1, 3, 7},
        {7, 6, 2, 0, 1, 5},
        {5, 1, 0, 2, 6, 7},
        {1, 0, 4, 6, 7, 3},
        {4, 0, 2, 3, 7, 5},
        {5, 4, 6, 2, 3, 1}};


    private final static float vertextable [] [] = {
        {-1.0f, -1.0f, -1.0f},
        {-1.0f, -1.0f,  1.0f},
        {-1.0f,  1.0f, -1.0f},
        {-1.0f,  1.0f,  1.0f},
        { 1.0f, -1.0f, -1.0f},
        { 1.0f, -1.0f,  1.0f},
        { 1.0f,  1.0f, -1.0f},
        { 1.0f,  1.0f,  1.0f}};





    CubeDrawer () {
        contourx = new float [6];
        contoury = new float [6];
        return;
    }




    void reset (Matrix3D mat, Matrix3D invmat, Dimension dim, float size,
                float xratio, float yratio, float zratio) {

        int center = 0;
        Vector3D vpoint = invmat.transform (0.0f, 0.0f, 100.0f);
        if (vpoint.x > -1.0f)   center += 4;
        if (vpoint.y > -1.0f)   center += 2;
        if (vpoint.z > -1.0f)   center += 1;

        Vector3D vertex;
        for (int i=0; i<6; i++) {
            int v = contourtable [center][i];
            vertex = mat.transform (
                vertextable [v][0] / xratio,
                vertextable [v][1] / yratio,
                vertextable [v][2] / zratio);

            contourx [i] =   (float) dim.width  * size * vertex.x;
            contoury [i] = - (float) dim.height * size * vertex.y;
            }


        vertex = mat.transform (
            vertextable [center][0] / xratio,
            vertextable [center][1] / yratio,
            vertextable [center][2] / zratio);

        centerx =   (float) dim.width  * size * vertex.x;
        centery = - (float) dim.height * size * vertex.y;
    }





    private int xpoints [] = new int [7];
    private int ypoints [] = new int [7];


    void draw (Graphics graph, Color color, float size, int x, int y) {
        for (int i=0; i<7; i++) {
            xpoints [i] = (int) (x + size * contourx [i%6]);
            ypoints [i] = (int) (y + size * contoury [i%6]);
            }

        int cenx = (int) (x + size * centerx);
        int ceny = (int) (y + size * centery);

        if (color != null)
            graph.setColor (color);
        else
            graph.setColor (Color.black);
        graph.fillPolygon (xpoints, ypoints, 6);


        if (color != null)
            graph.setColor (Color.black);
        else
            graph.setColor (Color.white);
        graph.drawPolygon (xpoints, ypoints, 7);
        graph.drawLine (cenx, ceny, xpoints[0], ypoints[0]);
        graph.drawLine (cenx, ceny, xpoints[2], ypoints[2]);
        graph.drawLine (cenx, ceny, xpoints[4], ypoints[4]);
    }


}




