import java.io.*;
import java.net.*;
import java.util.Vector;



class Record {
    private Object []  data;

    public Record (int size) {data = new Object [size];}
    public void   set (int field, Object value) {data [field] = value;}
    public Object get (int field) {return data [field];}
}




public class Database extends Object {

    public final static int INVALID = 0;
    public final static int NUMERIC = 1;
    public final static int STRING  = 2;

    public int        fieldcount;
    public String []  fieldname;
    public int    []  fieldtype;
    public int        reccount;
    public Vector     record;


    public Database (Database db, Filter filter) {
        fieldcount = db.fieldcount;
        int i, j;
        fieldname = new String [fieldcount];
        fieldtype = new int [fieldcount];
        for (i=0; i<db.fieldcount; i++) {
            fieldname [i] = new String (db.fieldname[i]);
            fieldtype [i] = db.fieldtype [i];
            }

        record = new Vector ();
        for (j=0; j<db.reccount; j++) {
            boolean out = false;
            for (i=0; i<filter.fcount; i++) {
                float value = ((Float) db.value
                              (j, filter.findex[i])).floatValue ();
                if ((filter.needmin[i] && value<filter.min[i]) ||
                    (filter.needmax[i] && value>filter.max[i])) {
                    out = true;
                    break;
                    }
                }


            if (! out) {
                Record rec = new Record (db.fieldcount);
                for (i=0; i<db.fieldcount; i++) {
                    Object objvalue;
                    if (db.fieldtype [i] == Database.STRING)
                        objvalue = new String ((String) db.value (j,i));
                    else
                        objvalue = new Float (((Float) db.value (j, i)).floatValue ());
                    rec.set (i, objvalue);
                    }
                record.addElement (rec);
                reccount ++;
                }
            }


    }




    public Database () throws Exception {
        reccount = 0;

            BufferedReader stream = new BufferedReader(new FileReader(System.getProperty("user.dir") + "\\3d.dat"));
            if (readFieldName (stream) == false)
                throw new IllegalArgumentException ();
            if (readFieldType (stream) == false)
                throw new IllegalArgumentException ();
            readData (stream);

        return;
    }



    private boolean readFieldName (BufferedReader stream) {

        String line;
        try {
            line = stream.readLine ();
        } catch (Exception e) {return false;}
        if (line == null)
            return false;


        int  len = line.length ();
        char [] buffer = new char [len];
        int  bufptr = 0;
        int  fieldnum = 0;
        boolean infield = false;
        Vector array = new Vector ();


        for (int i=0; i<=len; i++) {
            char ch;
            if (i < len)
                ch = line.charAt (i);
            else
                ch = ' ';

            if (infield == false) {
                if (ch != ' ' && ch != '\t') {
                    infield = true;
                    buffer [bufptr++] = ch;
                    }
                }
            else {
                if (ch != ' ' && ch != '\t')
                    buffer [bufptr++] = ch;
                else {
                    String name = new String (buffer, 0, bufptr);
                    array.addElement (name);
                    fieldnum ++;
                    bufptr = 0;
                    infield = false;
                    }
                }
            }


        if (fieldnum == 0)
            return false;

        fieldcount = fieldnum;
        fieldname = new String [fieldcount];
        for (int i=0; i<fieldnum; i++)
            fieldname [i] = (String) array.elementAt (i);
        return true;
    }





    private boolean readFieldType (BufferedReader stream) {

        String line;
        try {
            line = stream.readLine ();
        } catch (Exception e) {return false;}
        if (line == null)
            return false;


        int  len    = line.length ();
        char [] buffer = new char [len];
        int  bufptr = 0;
        int  fieldnum = 0;
        boolean infield = false;

        fieldtype = new int [fieldcount];
        for (int i=0; i<=len; i++) {
            char ch;
            if (i < len)
                ch = line.charAt (i);
            else
                ch = ' ';

            if (infield == false) {
                if (ch != ' ' && ch != '\t') {
                    infield = true;
                    buffer [bufptr++] = ch;
                    }
                }
            else {
                if (ch != ' ' && ch != '\t')
                    buffer [bufptr++] = ch;
                else {
                    if (fieldnum >= fieldcount)
                        return false;

                    String type = new String (buffer, 0, bufptr);
                    fieldtype [fieldnum] = INVALID;
                    if (type.equals ("String") ||
                        type.equals ("STRING") ||
                        type.equals ("string"))
                        fieldtype [fieldnum] = STRING;
                    if (type.equals ("Numeric") ||
                        type.equals ("NUMERIC") ||
                        type.equals ("numeric"))
                        fieldtype [fieldnum] = NUMERIC;

                    if (fieldtype [fieldnum] == INVALID)
                        return false;
                    fieldnum ++;
                    bufptr = 0;
                    infield = false;
                    }
                }
            }

        return true;
    }




    private void readData (BufferedReader stream) {
        reccount = 0;
        record = new Vector ();
        for (;;) {
            String line;
            try {
                line = stream.readLine ();
            } catch (Exception e) {return;}
            if (line == null)
                return;

            parseData (line);
            }
    }


    private static final int SYM_SPACE  = 1;
    private static final int SYM_NUMBER = 2;
    private static final int SYM_DOT    = 3;
    private static final int SYM_QUOTE  = 4;
    private static final int SYM_OTHERS = 5;

    private static final int ST_START   = 1;
    private static final int ST_PREDOT  = 2;
    private static final int ST_POSTDOT = 3;
    private static final int ST_INQUOTE = 4;




    private static int symbol (char ch) {
        if (ch==' ' || ch=='\t')
            return SYM_SPACE;
        else if (ch>='0' && ch<='9')
            return SYM_NUMBER;
        else if (ch == '.')
            return SYM_DOT;
        else if (ch == '"')
            return SYM_QUOTE;
        else
            return SYM_OTHERS;
    }






    private boolean parseData (String line) {

        int  len   = line.length ();
        char [] buffer = new char [len];
        int  bufptr = 0;
        int  state = ST_START;
        int  curfield = 0;
        int  dotcount = 10;
        boolean fielddone = false;
        boolean error = false;
        float   value = 0.0f;
        int     curtype = fieldtype [curfield];
        Record  rec = new Record (fieldcount);


        for (int i=0; i<=len; i++) {
            char ch;
            if (i < len)
                ch = line.charAt (i);
            else
                ch = ' ';
            int sym = symbol (ch);


            if (state == ST_START) {
                if (sym == SYM_SPACE) {
                    }
                else if (sym == SYM_NUMBER) {
                    value = (float) (ch-'0');
                    state = ST_PREDOT;
                    }
                else if (sym == SYM_DOT) {
                    value = 0.0f;
                    state = ST_POSTDOT;
                    dotcount = 10;
                    }
                else if (sym == SYM_QUOTE) {
                    bufptr = 0;
                    state = ST_INQUOTE;
                    }
                }


            else if (state == ST_PREDOT) {
                if (curtype != NUMERIC)
                    error = true;
                else if (sym == SYM_SPACE)
                    fielddone = true;
                else if (sym == SYM_NUMBER)
                    value = value*10 + (float)(ch-'0');
                else if (sym == SYM_DOT) {
                    state = ST_POSTDOT;
                    dotcount = 10;
                    }
                else if (sym == SYM_QUOTE || sym == SYM_OTHERS)
                    error = true;
                }


            else if (state == ST_POSTDOT) {
                if (curtype != NUMERIC)
                    error = true;
                else if (sym == SYM_SPACE)
                    fielddone = true;
                else if (sym == SYM_NUMBER) {
                    value += (float)(ch-'0') / (float)dotcount;
                    dotcount *= 10;
                    }
                else if (sym == SYM_DOT ||
                         sym == SYM_QUOTE ||
                         sym == SYM_OTHERS) {
                    error = true;
                    }
                }


            else if (state == ST_INQUOTE) {
                if (curtype != STRING)
                    error = true;
                else if (sym == SYM_QUOTE)
                    fielddone = true;
                else
                    buffer [bufptr++] = ch;
                }


            if (error)
                return false;

            if (fielddone) {
                Object content;
                if (curtype == STRING)
                    content = new String (buffer, 0, bufptr);
                else
                    content = new Float (value);
                rec.set (curfield++, content);
                fielddone = false;
                state = ST_START;
                if (curfield == fieldcount)
                    break;
                else
                    curtype = fieldtype [curfield];
                }
            }


        if (state != ST_START)
            return false;
        if (curfield == fieldcount) {
            record.addElement (rec);
            reccount ++;
            }
        return true;
    }




    public Object value (int recnum, int fieldnum) {
        return ((Record) record.elementAt (recnum)).get (fieldnum);
    }


    public float numericValue (int recnum, int fieldnum) {
        return ((Float) value (recnum, fieldnum)).floatValue ();
    }


    public String stringValue (int recnum, int fieldnum) {
        return (String) value (recnum, fieldnum);
    }
}

