import java.awt.*;



public class FilterBox extends Frame {

    int             fcount;
    int       []    findex;
    Label     []    label;
    Choice    []    mintext;
    Choice    []    maxtext;
    Panel     []    pmintext;
    Panel     []    pmaxtext;
    Label           error1;
    Label           error2;
    Button          ok;
    Button          apply;
    Button          cancel;
    Database        database;
    Plotter         plotter;
    
    

    public FilterBox (Plotter plotter, Database database) {
        super ("Database Filter");

        this.plotter = plotter;
        fcount = 0;
        findex = new int [database.fieldcount];
        for (int i=0; i<database.fieldcount; i++)
            if (database.fieldtype[i] == Database.NUMERIC)
                findex [fcount++] = i;

        label    = new Label  [fcount];
        mintext  = new Choice [fcount];
        maxtext  = new Choice [fcount];
        pmintext = new Panel  [fcount];
        pmaxtext = new Panel  [fcount];
        error1   = new Label ();
        error2   = new Label ();
        ok       = new Button ("OK");
        apply    = new Button ("Apply");
        cancel   = new Button ("Cancel");
        this.database = database;
        

        resize (300, fcount*40+140);
        setLayout (null);
        setBackground (Color.lightGray);
        

        for (int i=0; i<fcount; i++) {
            label[i] = new Label (database.fieldname [findex[i]]);
            pmintext[i] = new Panel ();
            pmaxtext[i] = new Panel ();
            mintext[i] = new Choice ();
            maxtext[i] = new Choice ();

            label[i].reshape (10, 45+i*40, 80, 30);
            pmintext[i].reshape (100, 45+i*40, 60, 30);
            pmaxtext[i].reshape (200, 45+i*40, 60, 30);
            
            setValues (i);
            
            add (label[i]);
            add (pmintext[i]);
            add (pmaxtext[i]);
            pmintext[i].add (mintext[i]);
            pmaxtext[i].add (maxtext[i]);
            }
            

        error1.reshape (20, 50+fcount*40, 240, 20);
        error2.reshape (20, 70+fcount*40, 240, 20);
        ok.reshape (40, 90+fcount*40, 50, 30);
        apply.reshape (120, 90+fcount*40, 50, 30);
        cancel.reshape (200, 90+fcount*40, 50, 30);
        add (ok);
        add (apply);
        add (cancel);
        add (error1);
        add (error2);
    }



    public synchronized boolean handleEvent (Event event) {
        if (event.id != Event.ACTION_EVENT)
            return false ;

        if ("Cancel".equals (event.arg)) {
            dispose ();
            plotter.active = true;
            }


        if ("OK".equals (event.arg) || "Apply".equals (event.arg)) {
            Filter filter = new Filter (fcount);

            for (int i=0; i<fcount; i++) {
                String minstr = mintext[i].getSelectedItem ();
                String maxstr = maxtext[i].getSelectedItem ();
                filter.needmin[i] = true;
                filter.needmax[i] = true;
                if (minstr.equals ("- N/A -"))
                    filter.needmin[i] = false;
                if (maxstr.equals ("- N/A -"))
                    filter.needmax[i] = false;
            
                filter.findex [i] = findex[i];
                filter.min [i] = 0.0f;
                filter.max [i] = 0.0f;
                
                if (filter.needmin[i])
                    filter.min[i] = Float.valueOf (minstr).floatValue ();
                if (filter.needmax[i])
                    filter.max[i] = Float.valueOf (maxstr).floatValue ();
                
                if (filter.needmin[i] &&
                    filter.needmax[i] &&
                    filter.min[i] >= filter.max[i]) {
                    error1.setText ("Min/Max mismatch for " + database.fieldname[findex[i]]);
                    return true;
                    }
                }
            
            Database newdatabase = new Database (database, filter);            
            plotter.loadData (newdatabase);
            plotter.setData (false);
            }
        
        
        if ("OK".equals (event.arg)) {
            plotter.active = true;
            dispose ();
            }

        return true;
    }



    private void setValues (int index) {
        float min;
        float max;
        min = max = database.numericValue (0, findex[index]);
        for (int i=1; i<database.reccount; i++) {
            float value = database.numericValue (i, findex[index]);
            if (value < min)
                min = value;
            if (value > max)
                max = value;
            }
            
        mintext[index].addItem ("- N/A -");
        maxtext[index].addItem ("- N/A -");
        for (int i=0; i<10; i++) {
            String str = String.valueOf (min+i*(max-min)/9.0f);
            mintext[index].addItem (str);
            maxtext[index].addItem (str);
            }
    }
}
