import java.awt.Color;
import java.util.Hashtable;



class Mapper extends Object {

    public int x;
    public int y;
    public int z;
    public int size;
    public int color;
    public Color defcolor;
    public Hashtable colorhash;

    public float minx;
    public float maxx;
    public float miny;
    public float maxy;
    public float minz;
    public float maxz;
    public float minsize;
    public float maxsize;



    public Mapper (int x, int y, int z,
                   int size, int color, Color defcolor,
                   float minx, float maxx,
                   float miny, float maxy,
                   float minz, float maxz,
                   float minsize, float maxsize,
                   Hashtable colorhash) {
        this.x = x;
        this.y = y;
        this.z = z;
        this.size = size;
        this.color = color;
        this.defcolor = defcolor;
        this.colorhash = colorhash;
        this.minx = minx;
        this.maxx = maxx;
        this.miny = miny;
        this.maxy = maxy;
        this.minz = minz;
        this.maxz = maxz;
        this.minsize = minsize;
        this.maxsize = maxsize;

        if (x < 0) {
            this.minx = -1.0f;
            this.maxx =  1.0f;
            }

        if (y < 0) {
            this.miny = -1.0f;
            this.maxy =  1.0f;
            }

        if (z < 0) {
            this.minz = -1.0f;
            this.maxz =  1.0f;
            }

        if (size < 0) {
            this.minsize = 1.0f;
            this.maxsize = 2.0f;
            }
    }



    public Mapper (Database database) {
        x = -1;
        y = -1;
        z = -1;
        color = -1;
        size  = -1;
        minx = -1.0f; maxx = 1.0f;
        miny = -1.0f; maxy = 1.0f;
        minz = -1.0f; maxz = 1.0f;
        minsize = 1.0f; maxsize = 2.0f;
        defcolor = Color.yellow;
        colorhash = null;

        int i;
        int state = 0;
        for (i=0; i<database.fieldcount; i++) {
            if (database.fieldtype [i] == Database.NUMERIC) {
                switch (state) {
                    case 0: x = i;  break;
                    case 1: y = i;  break;
                    case 2: z = i;  break;
                    case 3: size = i; break;
                    }
                state ++;
                }
            }
        
        rescale (database, true, true, true, true);
        return;
   }
   
   
   
   public void rescale (Database database,
                        boolean rescalex,
                        boolean rescaley,
                        boolean rescalez,
                        boolean rescalesize) {
        int i;
        boolean needx = x>=0 && rescalex;
        boolean needy = y>=0 && rescaley;
        boolean needz = z>=0 && rescalez;
        boolean needsize = size>=0 && rescalesize;

        if (needx)
            minx = maxx = ((Float) database.value (0, x)).floatValue ();

        if (needy)
            miny = maxy = ((Float) database.value (0, y)).floatValue ();

        if (needz)
            minz = maxz = ((Float) database.value (0, z)).floatValue ();

        if (needsize)
            minsize = maxsize = ((Float) database.value (0, size)).floatValue ();


        for (i=1; i<database.reccount; i++) {
            if (needx) {
                minx = Math.min (minx, ((Float) database.value (i, x)).floatValue ());
                maxx = Math.max (maxx, ((Float) database.value (i, x)).floatValue ());
                }

            if (needy) {
                miny = Math.min (miny, ((Float) database.value (i, y)).floatValue ());
                maxy = Math.max (maxy, ((Float) database.value (i, y)).floatValue ());
                }

            if (needz) {
                minz = Math.min (minz, ((Float) database.value (i, z)).floatValue ());
                maxz = Math.max (maxz, ((Float) database.value (i, z)).floatValue ());
                }

            if (needsize) {
                minsize = Math.min (minsize, ((Float) database.value (i, size)).floatValue ());
                maxsize = Math.max (maxsize, ((Float) database.value (i, size)).floatValue ());
                }
            }


    }




    public Color getcolor (String text) {
        if (color == -1)
            return defcolor;
        Color thecolor = (Color) colorhash.get (text);
        if (thecolor == null)
            return defcolor;
        else
            return thecolor;
    }


}


