
class Matrix3D extends Object {

    private float xx, xy, xz, xd;
    private float yx, yy, yz, yd;
    private float zx, zy, zz, zd;
    private float dx, dy, dz, dd;



    Matrix3D () {
        xx = 1.0f;
        yy = 1.0f;
        zz = 1.0f;
        dd = 1.0f;
    }


    Matrix3D (float mat [][]) {
        xx = mat[0][0]; xy = mat[0][1]; xz = mat[0][2]; xd = mat[0][3];
        yx = mat[1][0]; yy = mat[1][1]; yz = mat[1][2]; yd = mat[1][3];
        zx = mat[2][0]; zy = mat[2][1]; zz = mat[2][2]; zd = mat[2][3];
        dx = mat[3][0]; dy = mat[3][1]; dz = mat[3][2]; dd = mat[3][3];
    }


    Matrix3D (float mxx, float mxy, float mxz, float mxd,
              float myx, float myy, float myz, float myd,
              float mzx, float mzy, float mzz, float mzd,
              float mdx, float mdy, float mdz, float mdd) {

        xx = mxx;   xy = mxy;   xz = mxz;   xd = mxd;
        yx = myx;   yy = myy;   yz = myz;   yd = myd;
        zx = mzx;   zy = mzy;   zz = mzz;   zd = mzd;
        dx = mdx;   dy = mdy;   dz = mdz;   dd = mdd;
    }




    Matrix3D (Matrix3D m1, Matrix3D m2) {
        xx = m1.xx*m2.xx + m1.xy*m2.yx + m1.xz*m2.zx + m1.xd*m2.dx;
        xy = m1.xx*m2.xy + m1.xy*m2.yy + m1.xz*m2.zy + m1.xd*m2.dy;
        xz = m1.xx*m2.xz + m1.xy*m2.yz + m1.xz*m2.zz + m1.xd*m2.dz;
        xd = m1.xx*m2.xd + m1.xy*m2.yd + m1.xz*m2.zd + m1.xd*m2.dd;

        yx = m1.yx*m2.xx + m1.yy*m2.yx + m1.yz*m2.zx + m1.yd*m2.dx;
        yy = m1.yx*m2.xy + m1.yy*m2.yy + m1.yz*m2.zy + m1.yd*m2.dy;
        yz = m1.yx*m2.xz + m1.yy*m2.yz + m1.yz*m2.zz + m1.yd*m2.dz;
        yd = m1.yx*m2.xd + m1.yy*m2.yd + m1.yz*m2.zd + m1.yd*m2.dd;

        zx = m1.zx*m2.xx + m1.zy*m2.yx + m1.zz*m2.zx + m1.zd*m2.dx;
        zy = m1.zx*m2.xy + m1.zy*m2.yy + m1.zz*m2.zy + m1.zd*m2.dy;
        zz = m1.zx*m2.xz + m1.zy*m2.yz + m1.zz*m2.zz + m1.zd*m2.dz;
        zd = m1.zx*m2.xd + m1.zy*m2.yd + m1.zz*m2.zd + m1.zd*m2.dd;

        dx = m1.dx*m2.xx + m1.dy*m2.yx + m1.dz*m2.zx + m1.dd*m2.dx;
        dy = m1.dx*m2.xy + m1.dy*m2.yy + m1.dz*m2.zy + m1.dd*m2.dy;
        dz = m1.dx*m2.xz + m1.dy*m2.yz + m1.dz*m2.zz + m1.dd*m2.dz;
        dd = m1.dx*m2.xd + m1.dy*m2.yd + m1.dz*m2.zd + m1.dd*m2.dd;
    }



    void set (float mxx, float mxy, float mxz, float mxd,
              float myx, float myy, float myz, float myd,
              float mzx, float mzy, float mzz, float mzd,
              float mdx, float mdy, float mdz, float mdd) {

        xx = mxx;   xy = mxy;   xz = mxz;   xd = mxd;
        yx = myx;   yy = myy;   yz = myz;   yd = myd;
        zx = mzx;   zy = mzy;   zz = mzz;   zd = mzd;
        dx = mdx;   dy = mdy;   dz = mdz;   dd = mdd;
    }



    void set (float mat [][]) {
        xx = mat[0][0]; xy = mat[0][1]; xz = mat[0][2]; xd = mat[0][3];
        yx = mat[1][0]; yy = mat[1][1]; yz = mat[1][2]; yd = mat[1][3];
        zx = mat[2][0]; zy = mat[2][1]; zz = mat[2][2]; zd = mat[2][3];
        dx = mat[3][0]; dy = mat[3][1]; dz = mat[3][2]; dd = mat[3][3];
    }




    void preScale (float factorx, float factory, float factorz) {
        xx *= factorx ;
        yx *= factorx ;
        zx *= factorx ;
        dx *= factorx ;
        xy *= factory ;
        yy *= factory ;
        zy *= factory ;
        dy *= factory ;
        xz *= factorz ;
        yz *= factorz ;
        zz *= factorz ;
        dz *= factorz ;
    }



    void postScale (float factorx, float factory, float factorz) {
        xx *= factorx ;
        xy *= factorx ;
        xz *= factorx ;
        xd *= factorx ;
        yx *= factory ;
        yy *= factory ;
        yz *= factory ;
        yd *= factory ;
        zx *= factorz ;
        zy *= factorz ;
        zz *= factorz ;
        zd *= factorz ;
    }


    void preTranslate (float x, float y, float z) {
        xd += x * xx ;
        yd += y * yy ;
        zd += z * zz ;
    }



    void preTranslate (Vector3D v) {
        xd += v.x * xx ;
        yd += v.y * yy ;
        zd += v.z * zz ;
    }



    void postTranslate (float x, float y, float z) {
        xx += x * dx ;
        xy += x * dy ;
        xz += x * dz ;
        xd += x * dd ;
        yx += y * dx ;
        yy += y * dy ;
        yz += y * dz ;
        yd += y * dd ;
        zx += z * dx ;
        zy += z * dy ;
        zz += z * dz ;
        zd += z * dd ;
    }




    void postTranslate (Vector3D v) {
        float x = v.x;
        float y = v.y;
        float z = v.z;

        xx += x * dx ;
        xy += x * dy ;
        xz += x * dz ;
        xd += x * dd ;
        yx += y * dx ;
        yy += y * dy ;
        yz += y * dz ;
        yd += y * dd ;
        zx += z * dx ;
        zy += z * dy ;
        zz += z * dz ;
        zd += z * dd ;
    }




    void preRotateX (double theta) {
        double cost = Math.cos (theta);
        double sint = Math.sin (theta);

        float nxy = (float) (xy * cost + xz * sint) ;
        float nyy = (float) (yy * cost + yz * sint) ;
        float nzy = (float) (zy * cost + zz * sint) ;
        float ndy = (float) (dy * cost + dz * sint) ;

        float nxz = (float) (xz * cost - xy * sint) ;
        float nyz = (float) (yz * cost - yy * sint) ;
        float nzz = (float) (zz * cost - zy * sint) ;
        float ndz = (float) (dz * cost - dy * sint) ;

        xy = nxy ;
        yy = nyy ;
        zy = nzy ;
        dy = ndy ;
        xz = nxz ;
        yz = nyz ;
        zz = nzz ;
        dz = ndz ;
    }





    void preRotateY (double theta) {
        double cost = Math.cos (theta);
        double sint = Math.sin (theta);

        float nxz = (float) (xz * cost + xx * sint) ;
        float nyz = (float) (yz * cost + yx * sint) ;
        float nzz = (float) (zz * cost + zx * sint) ;
        float ndz = (float) (dz * cost + dx * sint) ;

        float nxx = (float) (xx * cost - xz * sint) ;
        float nyx = (float) (yx * cost - yz * sint) ;
        float nzx = (float) (zx * cost - zz * sint) ;
        float ndx = (float) (dx * cost - dz * sint) ;

        xz = nxz ;
        yz = nyz ;
        zz = nzz ;
        dz = ndz ;
        xx = nxx ;
        yx = nyx ;
        zx = nzx ;
        dx = ndx ;
    }




    void preRotateZ (double theta) {
        double cost = Math.cos (theta);
        double sint = Math.sin (theta);

        float nxx = (float) (xx * cost + xy * sint) ;
        float nyx = (float) (yx * cost + yy * sint) ;
        float nzx = (float) (zx * cost + zy * sint) ;
        float ndx = (float) (dx * cost + dy * sint) ;

        float nxy = (float) (xy * cost - xx * sint) ;
        float nyy = (float) (yy * cost - yx * sint) ;
        float nzy = (float) (zy * cost - zx * sint) ;
        float ndy = (float) (dy * cost - dx * sint) ;

        xx = nxx ;
        yx = nyx ;
        zx = nzx ;
        dx = ndx ;
        xy = nxy ;
        yy = nyy ;
        zy = nzy ;
        dy = ndy ;
    }






    void postRotateX (double theta) {
        double cost = Math.cos (theta);
        double sint = Math.sin (theta);

        float nyx = (float) (yx * cost - zx * sint) ;
        float nyy = (float) (yy * cost - zy * sint) ;
        float nyz = (float) (yz * cost - zz * sint) ;
        float nyd = (float) (yd * cost - zd * sint) ;

        float nzx = (float) (zx * cost + yx * sint) ;
        float nzy = (float) (zy * cost + yy * sint) ;
        float nzz = (float) (zz * cost + yz * sint) ;
        float nzd = (float) (zd * cost + yd * sint) ;

        yx = nyx ;
        yy = nyy ;
        yz = nyz ;
        yd = nyd ;
        zx = nzx ;
        zy = nzy ;
        zz = nzz ;
        zd = nzd ;
    }



    void postRotateY (double theta) {
        double cost = Math.cos (theta);
        double sint = Math.sin (theta);

        float nzx = (float) (zx * cost - xx * sint) ;
        float nzy = (float) (zy * cost - xy * sint) ;
        float nzz = (float) (zz * cost - xz * sint) ;
        float nzd = (float) (zd * cost - xd * sint) ;

        float nxx = (float) (xx * cost + zx * sint) ;
        float nxy = (float) (xy * cost + zy * sint) ;
        float nxz = (float) (xz * cost + zz * sint) ;
        float nxd = (float) (xd * cost + zd * sint) ;

        zx = nzx ;
        zy = nzy ;
        zz = nzz ;
        zd = nzd ;
        xx = nxx ;
        xy = nxy ;
        xz = nxz ;
        xd = nxd ;
    }




    void postRotateZ (double theta) {
        double cost = Math.cos (theta);
        double sint = Math.sin (theta);

        float nxx = (float) (xx * cost - yx * sint) ;
        float nxy = (float) (xy * cost - yy * sint) ;
        float nxz = (float) (xz * cost - yz * sint) ;
        float nxd = (float) (xd * cost - yd * sint) ;

        float nyx = (float) (yx * cost + xx * sint) ;
        float nyy = (float) (yy * cost + xy * sint) ;
        float nyz = (float) (yz * cost + xz * sint) ;
        float nyd = (float) (yd * cost + xd * sint) ;

        xx = nxx ;
        xy = nxy ;
        xz = nxz ;
        xd = nxd ;
        yx = nyx ;
        yy = nyy ;
        yz = nyz ;
        yd = nyd ;
    }




    void unify () {
        xx = 1;   xy = 0;   xz = 0;   xd = 0;
        yx = 0;   yy = 1;   yz = 0;   yd = 0;
        zx = 0;   zy = 0;   zz = 1;   zd = 0;
        dx = 0;   dy = 0;   dz = 0;   dd = 1;
    }



    Vector3D transform (Vector3D v) {
        float x = v.x;
        float y = v.y;
        float z = v.z;

        float d = x*dx + y*dy + z*dz + dd;
        return new Vector3D ((x*xx + y*xy + z*xz + xd) / d,
                             (x*yx + y*yy + z*yz + yd) / d,
                             (x*zx + y*zy + z*zz + zd) / d);
    }




    Vector3D transform (float x, float y, float z) {
        float d = x*dx + y*dy + z*dz + dd;
        return new Vector3D ((x*xx + y*xy + z*xz + xd) / d,
                             (x*yx + y*yy + z*yz + yd) / d,
                             (x*zx + y*zy + z*zz + zd) / d);
    }





    Matrix3D inverse () {
        float A11, A12, A13, A14;
        float A21, A22, A23, A24;
        float A31, A32, A33, A34;
        float A41, A42, A43, A44;

        A11 =   yy*zz*dd + zy*dz*yd + dy*yz*zd - yd*zz*dy - zd*dz*yy - dd*yz*zy ;
        A12 = -(yx*zz*dd + zx*dz*yd + dx*yz*zd - yd*zz*dx - zd*dz*yx - dd*yz*zx);
        A13 =   yx*zy*dd + zx*dy*yd + dx*yy*zd - yd*zy*dx - zd*dy*yx - dd*yy*zx ;
        A14 = -(yx*zy*dz + zx*dy*yz + dx*yy*zz - yz*zy*dx - zz*dy*yx - dz*yy*zx);

        A21 = -(xy*zz*dd + zy*dz*xd + dy*xz*zd - xd*zz*dy - zd*dz*xy - dd*xz*zy);
        A22 =   xx*zz*dd + zx*dz*xd + dx*xz*zd - xd*zz*dx - zd*dz*xx - dd*xz*zx ;
        A23 = -(xx*zy*dd + zx*dy*xd + dx*xy*zd - xd*zy*dx - zd*dy*xx - dd*xy*zx);
        A24 =   xx*zy*dz + zx*dy*xz + dx*xy*zz - xz*zy*dx - zz*dy*xx - dz*xy*zx ;

        A31 =   xy*yz*dd + yy*dz*xd + dy*xz*yd - xd*yz*dy - yd*dz*xy - dd*xz*yy ;
        A32 = -(xx*yz*dd + yx*dz*xd + dx*xz*yd - xd*yz*dx - yd*dz*xx - dd*xz*yx);
        A33 =   xx*yy*dd + yx*dy*xd + dx*xy*yd - xd*yy*dx - yd*dy*xx - dd*xy*yx ;
        A34 = -(xx*yy*dz + yx*dy*xz + dx*xy*yz - xz*yy*dx - yz*dy*xx - dz*xy*yx);

        A41 = -(xy*yz*zd + yy*zz*xd + zy*xz*yd - xd*yz*zy - yd*zz*xy - zd*xz*yy);
        A42 =   xx*yz*zd + yx*zz*xd + zx*xz*yd - xd*yz*zx - yd*zz*xx - zd*xz*yx ;
        A43 = -(xx*yy*zd + yx*zy*xd + zx*xy*yd - xd*yy*zx - yd*zy*xx - zd*xy*yx);
        A44 =   xx*yy*zz + yx*zy*xz + zx*xy*yz - xz*yy*zx - yz*zy*xx - zz*xy*yx ;

        float det = xx*A11 + xy*A12 + xz*A13 + xd*A14;

        float mat [][] = new float [4][4];
        mat[0][0]=A11/det; mat[0][1]=A21/det; mat[0][2]=A31/det; mat[0][3]=A41/det;
        mat[1][0]=A12/det; mat[1][1]=A22/det; mat[1][2]=A32/det; mat[1][3]=A42/det;
        mat[2][0]=A13/det; mat[2][1]=A23/det; mat[2][2]=A33/det; mat[2][3]=A43/det;
        mat[3][0]=A14/det; mat[3][1]=A24/det; mat[3][2]=A34/det; mat[3][3]=A44/det;

        return new Matrix3D (mat);
    }





    public void print () {
        System.out.println ("--------------------");
        System.out.println ("|" + xx + " " + xy + " " + xz + " " + xd + "| ");
        System.out.println ("|" + yx + " " + yy + " " + yz + " " + yd + "| ");
        System.out.println ("|" + zx + " " + zy + " " + zz + " " + zd + "| ");
        System.out.println ("|" + dx + " " + dy + " " + dz + " " + dd + "| ");
    }
}




