import java.awt.*;


public class PlotView extends Canvas implements WheelClient, ButtonClient, CheckerClient {

    Plotter plotter;

    PlotData   data [];
    PlotData   vdata [];
    public Mapper     mapper;
    int           datacount;
    float         minx, maxx;
    float         miny, maxy;
    float         minz, maxz;
    float         minsize, maxsize;
    int           selectid;

    String        xlabel;
    String        ylabel;
    String        zlabel;
    boolean       needgrid;
    boolean       needreport;
    int           reportx;
    int           reporty;
    boolean       opaqueplane;
    float         bulletsize;
    float         xinit;
    float         yinit;
    float         zinit;
    float         xratio;
    float         yratio;
    float         zratio;
    float         zoomer;
    double        viewangle;
    Vector3D      viewpoint;
    Matrix3D      viewmat;
    Matrix3D      deltamat;
    Matrix3D      transmat;
    Matrix3D      invmat;


    CubeDrawer    cubedrawer;
    Image         bullet;
    Image         offscreen;
    Graphics      offgraphics;
    Dimension     offdim;
    Font          reportfont = new Font ("Helvetica", Font.PLAIN, 10);


    boolean       animode;
    int           curstep;
    float         alpha;
    PlotData      backupdata [];
    int           backupsize;
    int           pathcount;
    int []        path;
    boolean       animatex;
    boolean       animatey;
    boolean       animatez;
    boolean       animatesize;



    PlotView (Plotter plotter, int width, int height) {
        resize (width, height);

        animode = false;

        this.plotter = plotter;
        datacount  = 0;

        minx = -1.0f;   maxx = 1.0f;
        miny = -1.0f;   maxy = 1.0f;
        minz = -1.0f;   maxz = 1.0f;

//        bullet = Frame3D.ireader.getimage ("bullet.gif");
        cubedrawer  = new CubeDrawer ();
        offscreen   = null;
        offgraphics = null;
        offdim      = size ();

        needgrid    = false;
        opaqueplane = false;
        bulletsize  = 0.075f;
        zoomer      = 0.25f;
        xinit       = -1.2f;
        yinit       = -1.2f;
        zinit       = -1.2f;
        xratio      = 1.0f;
        yratio      = 1.0f;
        zratio      = 1.0f;
        viewpoint   = new Vector3D (1.0f, 1.0f, 1.0f);
        viewangle   = Math.PI;
        deltamat    = new Matrix3D ();
        viewmat     = new Matrix3D ();
        selectid    = -1;
        needreport  = false;
        setView ();
    }



    public void setView () {

        float x = viewpoint.x;
        float y = viewpoint.y;
        float z = viewpoint.z;

        float cospsi;
        float sinpsi;
        float costheta;
        float sintheta;
        float cosphi;
        float sinphi;


        if (Math.abs (x) > 0.001 || Math.abs (y) > 0.001) {
            cospsi   = (float) Math.cos (viewangle);
            sinpsi   = (float) Math.sin (viewangle);
            costheta = y / (float) Math.sqrt (x*x + y*y);
            sintheta = x / (float) Math.sqrt (x*x + y*y);
            cosphi   = z / (float) Math.sqrt (x*x + y*y + z*z);
            sinphi   = (float) (Math.sqrt (x*x + y*y) /
                       Math.sqrt (x*x + y*y + z*z));
            }
        else {
            cospsi   = (float) Math.cos (viewangle);
            sinpsi   = (float) Math.sin (viewangle);
            costheta = 1.0f;
            sintheta = 0.0f;
            cosphi   = z / (float) Math.sqrt (x*x + y*y + z*z);
            sinphi   = 0.0f;
            }


/*
        float matxx = xratio * ( cospsi*costheta + sinpsi*sintheta);
        float matxy = yratio * (-cospsi*sintheta - sinpsi*cosphi*costheta);
        float matxz = zratio * ( sinpsi*sinphi);

        float matyx = xratio * ( sinpsi*costheta + cospsi*cosphi*sintheta);
        float matyy = yratio * (-sinpsi*sintheta + cospsi*cosphi*costheta);
        float matyz = zratio * (-cospsi*sinphi);

        float matzx = xratio * sinphi * sintheta;
        float matzy = yratio * sinphi * costheta;
        float matzz = zratio * cosphi;


        viewmat.set (zoomer*matxx, zoomer*matxy, zoomer*matxz, 0.0f,
                     zoomer*matyx, zoomer*matyy, zoomer*matyz, 0.0f,
                     zoomer*matzx, zoomer*matzy, zoomer*matzz, 0.0f,
                             0.0f,         0.0f,         0.0f, 1.0f);
*/

        float [] [] mat = new float [4][4];

        mat[0][0] = zoomer * xratio * ( cospsi*costheta + sinpsi*sintheta);
        mat[0][1] = zoomer * yratio * (-cospsi*sintheta - sinpsi*cosphi*costheta);
        mat[0][2] = zoomer * zratio * ( sinpsi*sinphi);
        mat[0][3] = 0.0f;

        mat[1][0] = zoomer * xratio * ( sinpsi*costheta + cospsi*cosphi*sintheta);
        mat[1][1] = zoomer * yratio * (-sinpsi*sintheta + cospsi*cosphi*costheta);
        mat[1][2] = zoomer * zratio * (-cospsi*sinphi);
        mat[1][3] = 0.0f;

        mat[2][0] = zoomer * xratio * sinphi * sintheta;
        mat[2][1] = zoomer * yratio * sinphi * costheta;
        mat[2][2] = zoomer * zratio * cosphi;
        mat[2][3] = 0.0f;

        mat[3][0] = 0.0f;
        mat[3][1] = 0.0f;
        mat[3][2] = 0.0f;
        mat[3][3] = 1.0f;


        viewmat.set (mat);
    }




    public boolean setData (Database database, Mapper mapper, boolean rescale) {
        if (database.reccount <= 0)
            return false;

        datacount = database.reccount;
        data = new PlotData [datacount];
        vdata = new PlotData [datacount];
        this.mapper = mapper;


        if (mapper.x >= 0)
            xlabel = database.fieldname [mapper.x];
        else
            xlabel = new String ();

        if (mapper.y >= 0)
            ylabel = database.fieldname [mapper.y];
        else
            ylabel = new String ();

        if (mapper.z >= 0)
            zlabel = database.fieldname [mapper.z];
        else
            zlabel = new String ();

        String sizelabel;
        if (mapper.size >= 0)
            sizelabel = database.fieldname [mapper.size];
        else
            sizelabel = new String ();

        if (rescale) {
                minx = mapper.minx;
                maxx = mapper.maxx;
                miny = mapper.miny;
                maxy = mapper.maxy;
                minz = mapper.minz;
                maxz = mapper.maxz;
                minsize = mapper.minsize;
                maxsize = mapper.maxsize;
            }



        for (int i=0; i<datacount; i++) {
            float x, y, z;
            float size;
            Color color;

            if (mapper.x==-1)
                x = 0.0f;
            else
                x = ((Float) database.value (i,mapper.x)).floatValue ();
            if (mapper.y==-1)
                y = 0.0f;
            else
                y = ((Float) database.value (i,mapper.y)).floatValue ();
            if (mapper.z==-1)
                z = 0.0f;
            else
                z = ((Float) database.value (i,mapper.z)).floatValue ();
            if (mapper.size==-1)
                size = 1.0f;
            else
                size = ((Float) database.value (i,mapper.size)).floatValue ();


            if (mapper.color == -1)
                color = mapper.defcolor;
            else
                color = mapper.getcolor (
                    (String) database.value (i, mapper.color));


            data [i] = new PlotData (
                i, color,
                1.0f + 1.0f * (size-minsize) / (maxsize-minsize),
                (2.0f * x - (minx+maxx)) / (maxx - minx),
                (2.0f * y - (miny+maxy)) / (maxy - miny),
                (2.0f * z - (minz+maxz)) / (maxz - minz));
            }

        LegendBox.setScale (xlabel, minx, maxx,
                            ylabel, miny, maxy,
                            zlabel, minz, maxz,
                            sizelabel);

        selectid = -1;
        repaint ();
        return true;
    }



    private void select (int x, int y) {
        int w = size ().width;
        int h = size ().height;
        int choice = -1;

        for (int i=0; i<datacount; i++) {
            int deltax = x - (int) (w/2 + w * vdata[i].viewvec.x);
            int deltay = y - (int) (h/2 - h * vdata[i].viewvec.y);
            if (((deltax>=0 && deltax<=5) ||
                 (deltax<=0 && deltax>=-5)) &&
                ((deltay>=0 && deltay<=5) ||
                 (deltay<=0 && deltay>=-5)))
                  choice = vdata[i].id;
            }


        if (selectid == choice)
            selectid = -1;
        else
            selectid = choice;
        if (selectid != -1) {
            needreport = true;
            reportx = x;
            reporty = y;
            }

        repaint ();
    }





    private void drawGrid (Graphics graph, Color axiscolor, Vector3D xaxis, Vector3D yaxis, String label) {

        int w = size ().width;
        int h = size ().height;


        Vector3D ovec = transmat.transform (xinit, yinit, zinit);
        Vector3D xvec = transmat.transform (xaxis.x * (1.0f-xinit) + xinit,
                                            xaxis.y * (1.0f-yinit) + yinit,
                                            xaxis.z * (1.0f-zinit) + zinit);
        Vector3D yvec = transmat.transform (yaxis.x * (1.0f-xinit) + xinit,
                                            yaxis.y * (1.0f-yinit) + yinit,
                                            yaxis.z * (1.0f-zinit) + zinit);
        Vector3D dvec = transmat.transform ((xaxis.x+yaxis.x) * (1.0f-xinit) + xinit,
                                            (xaxis.y+yaxis.y) * (1.0f-yinit) + yinit,
                                            (xaxis.z+yaxis.z) * (1.0f-zinit) + zinit);

        if (opaqueplane) {
            int xpoints [] = new int [5];
            int ypoints [] = new int [5];

            xpoints [0] = (int) (w/2 + w * ovec.x);
            xpoints [1] = (int) (w/2 + w * xvec.x);
            xpoints [2] = (int) (w/2 + w * dvec.x);
            xpoints [3] = (int) (w/2 + w * yvec.x);
            xpoints [4] = xpoints [0];

            ypoints [0] = (int) (h/2 - h * ovec.y);
            ypoints [1] = (int) (h/2 - h * xvec.y);
            ypoints [2] = (int) (h/2 - h * dvec.y);
            ypoints [3] = (int) (h/2 - h * yvec.y);
            ypoints [4] = ypoints [0];


            graph.setColor (Color.gray);
            graph.fillPolygon (xpoints, ypoints, 4);
            graph.setColor (Color.black);
            graph.drawPolygon (xpoints, ypoints, 4);
            }
        else {
            ovec.z = 0.0f;
            xvec.z = 0.0f;

            int ox = (int) (w/2 + w * ovec.x);
            int xx = (int) (w/2 + w * xvec.x);
            int oy = (int) (h/2 - h * ovec.y);
            int xy = (int) (h/2 - h * xvec.y);

            int posdeltax [] = {0, 0, 0, 1, 1};
            int posdeltay [] = {-1, -1, 0, 0, 0};
            int negdeltax [] = {0, 0, 0, -1, -1};
            int negdeltay [] = {1, 1, 0, 0, 0};
            int deltax [];
            int deltay [];

            if (xx > ox)
                deltax = posdeltax;
            else
                deltax = negdeltax;
            if (xy > oy)
                deltay = posdeltay;
            else
                deltay = negdeltay;

            graph.setColor (Color.black);
            for (int i=0; i<5; i++)
                graph.drawLine (ox+deltax[i], oy+deltay[i],
                                xx+deltax[i], xy+deltay[i]);

            Vector3D xov  = xvec.minus (ovec);
            if (xov.length () > 0.001) {
                Vector3D kvec = xov.normalize (0.03f);
                Vector3D hvec = new Vector3D (-kvec.y, kvec.x, 0.0f).normalize (0.01f);
                Vector3D pvec = (xvec.minus (kvec)).plus  (hvec);
                Vector3D qvec = (xvec.minus (kvec)).minus (hvec);


                int xpoints [] = new int [4];
                int ypoints [] = new int [4];

                xpoints [0] = (int) (w/2 + w * xvec.x);
                xpoints [1] = (int) (w/2 + w * pvec.x);
                xpoints [2] = (int) (w/2 + w * qvec.x);
                xpoints [3] = xpoints [0];

                ypoints [0] = (int) (h/2 - h * xvec.y);
                ypoints [1] = (int) (h/2 - h * pvec.y);
                ypoints [2] = (int) (h/2 - h * qvec.y);
                ypoints [3] = ypoints [0];


                int xtext = (xpoints[0] + (xpoints[1] + xpoints[2])/2)/2 + 6;
                int ytext = (ypoints[0] + (ypoints[1] + ypoints[2])/2)/2 + 4;

                graph.setColor (Color.black);
                graph.fillPolygon (xpoints, ypoints, 3);
                graph.drawString (label, xtext, ytext);
                }
            }


        if (needgrid) {
            graph.setColor (Color.white);
            for (int i=1; i<=9; i++) {
                graph.drawLine (
                    (int) (w/2 + w * (ovec.x + ((float)i/10.0f) * (xvec.x-ovec.x))),
                    (int) (h/2 - h * (ovec.y + ((float)i/10.0f) * (xvec.y-ovec.y))),
                    (int) (w/2 + w * (yvec.x + ((float)i/10.0f) * (dvec.x-yvec.x))),
                    (int) (h/2 - h * (yvec.y + ((float)i/10.0f) * (dvec.y-yvec.y))));
                graph.drawLine (
                    (int) (w/2 + w * (ovec.x + ((float)i/10.0f) * (yvec.x-ovec.x))),
                    (int) (h/2 - h * (ovec.y + ((float)i/10.0f) * (yvec.y-ovec.y))),
                    (int) (w/2 + w * (xvec.x + ((float)i/10.0f) * (dvec.x-xvec.x))),
                    (int) (h/2 - h * (xvec.y + ((float)i/10.0f) * (dvec.y-xvec.y))));
                }
            }

    }





    private void animateData () {

        for (int i=0; i<curstep; i++) {
            backupdata[i] = new PlotData (data[path[i]]);
            if (animatesize)
                data[path[i]].size = data[path[i+1]].size;
            if (animatex)
                data[path[i]].vec.x = data[path[i+1]].vec.x;
            if (animatey)
                data[path[i]].vec.y = data[path[i+1]].vec.y;
            if (animatez)
                data[path[i]].vec.z = data[path[i+1]].vec.z;
            }

        backupdata[curstep] = new PlotData (data[path[curstep]]);
        backupsize = curstep+1;
        PlotData thisdata = data [path[curstep]];
        PlotData nextdata = data [path[curstep+1]];


        if (animatesize)
            thisdata.size = thisdata.size + alpha * (nextdata.size-thisdata.size);
        if (animatex)
            thisdata.vec.x = thisdata.vec.x + alpha * (nextdata.vec.x-thisdata.vec.x);
        if (animatey)
            thisdata.vec.y = thisdata.vec.y + alpha * (nextdata.vec.y-thisdata.vec.y);
        if (animatez)
            thisdata.vec.z = thisdata.vec.z + alpha * (nextdata.vec.z-thisdata.vec.z);
    }



    private void restoreData () {
        for (int i=0; i<backupsize; i++)
            data[path[i]] = new PlotData (backupdata[i]);
    }



    private void drawData (Graphics graph) {

        int w = size ().width;
        int h = size ().height;
        int i;


        Vector3D vpoint = invmat.transform (0.0f, 0.0f, 1000.0f);

        if (vpoint.x + 1.0f > xinit)
            drawGrid (graph, Color.red,
                      new Vector3D (0.0f, 1.0f, 0.0f),
                      new Vector3D (0.0f, 0.0f, 1.0f),
                      ylabel);

        if (vpoint.y + 1.0f > yinit)
            drawGrid (graph, Color.blue,
                      new Vector3D (0.0f, 0.0f, 1.0f),
                      new Vector3D (1.0f, 0.0f, 0.0f),
                      zlabel);

        if (vpoint.z + 1.0f > zinit)
            drawGrid (graph, Color.green,
                      new Vector3D (1.0f, 0.0f, 0.0f),
                      new Vector3D (0.0f, 1.0f, 0.0f),
                      xlabel);


        if (animode)
            animateData ();


        for (i=0; i<datacount; i++) {
            vdata[i] = new PlotData (data[i]);
            vdata[i].viewvec = transmat.transform (vdata[i].vec);
            }
        if (datacount >= 1)
            quicksort (0, datacount-1);


        for (i=0; i<datacount; i++) {
            Color color;
            if (vdata[i].id == selectid)
                color = null;
            else
                color = vdata[i].color;

            cubedrawer.draw (graph, color, vdata[i].size,
                (int) (w/2 + w * vdata[i].viewvec.x),
                (int) (h/2 - h * vdata[i].viewvec.y));
            }


        if (animode)
            restoreData ();



        if (vpoint.x + 1.0f < xinit)
            drawGrid (graph, Color.red,
                      new Vector3D (0.0f, 1.0f, 0.0f),
                      new Vector3D (0.0f, 0.0f, 1.0f),
                      ylabel);

        if (vpoint.y + 1.0f < yinit)
            drawGrid (graph, Color.blue,
                      new Vector3D (0.0f, 0.0f, 1.0f),
                      new Vector3D (1.0f, 0.0f, 0.0f),
                      zlabel);

        if (vpoint.z + 1.0f < zinit)
            drawGrid (graph, Color.green,
                      new Vector3D (1.0f, 0.0f, 0.0f),
                      new Vector3D (0.0f, 1.0f, 0.0f),
                      xlabel);

    }



    public void paint (Graphics graph) {
        update (graph);
        return;
    }



    public void report (Graphics graph, Database database) {
        int w = size ().width;
        int h = size ().height;
        int fcount = database.fieldcount;
        int lineheight = 15;

        int y = reporty + 5;
        if (y + lineheight*fcount + 8>= h)
            y = h - lineheight*fcount - 8;

        int x = reportx;
        if (x + 152 >= w)
            x = w - 152;

        Graphics clipgraph = graph.create (x, y, 152, lineheight*fcount+8);
        clipgraph.setColor (Color.white);
        clipgraph.fillRect (0, 0, 150, 6+lineheight*fcount);
        clipgraph.setColor (Color.black);
        clipgraph.drawRect (0, 0, 150, 6+lineheight*fcount);
        clipgraph.setFont  (reportfont);

        for (int i=0; i<fcount; i++) {
            clipgraph.drawString (database.fieldname [i], 5, (i+1)*lineheight);
            String string = database.value (selectid, i).toString ();
            clipgraph.drawString (string, 75, (i+1)*lineheight);
            }
    }


    public synchronized void update (Graphics graph) {

        Dimension dim = size ();
        if ((offscreen == null) ||
            (dim.width  != offdim.width) ||
            (dim.height != offdim.height)) {
            offscreen = createImage (dim.width, dim.height);
            offdim = dim;
            offgraphics = offscreen.getGraphics();
            }


        offgraphics.setColor (Color.lightGray) ;
        offgraphics.fillRect (0, 0, dim.width, dim.height);
        transmat = new Matrix3D (deltamat, viewmat);
        invmat   = transmat.inverse ();

        cubedrawer.reset (transmat, invmat, size (), bulletsize,
                          xratio, yratio, zratio);


        drawData (offgraphics);
        if (needreport && ! animode) {
            report (offgraphics, plotter.database);
            needreport = false;
            }

        graph.drawImage (offscreen, 0, 0, null);
    }




    private void quicksort (int left, int right) {
        int i, j;
        int mid;
        PlotData tmp;

        i   = left;
        j   = right;
        mid = (left+right) / 2;

        do {
            while (vdata [mid].viewvec.z > vdata [i].viewvec.z && i < right)
                i ++;
            while (vdata [j].viewvec.z > vdata [mid].viewvec.z && j > left)
                j --;

            if (i <= j) {
                PlotData tempdata ;
                tempdata = vdata [i];
                vdata [i] = vdata [j];
                vdata [j] = tempdata;
                i ++ ;
                j -- ;
            }

        } while (i <= j);

        if (left < j)
            quicksort (left, j);
        if (i < right)
            quicksort (i, right);
    }




    int      prevx;
    int      prevy;
    int      downx;
    int      downy;
    boolean  mousedown = false;


    public boolean mouseDown (Event event, int x, int y) {
        prevx = downx = x;
        prevy = downy = y;
        mousedown = true;
        return true;
    }


    public boolean mouseUp (Event event, int x, int y) {
        mousedown = false;
        if (x==downx && y==downy)
            select (x, y);
        return true;
    }



    public boolean mouseDrag (Event event, int x, int y) {
        if (mousedown == false)
            return true;

        double xrot = (y - prevy) * 2 * Math.PI / size().height;
        double yrot = (x - prevx) * 2 * Math.PI / size().width;
        deltamat.postRotateX (xrot);
        deltamat.postRotateY (yrot);
        prevx = x;
        prevy = y;

        repaint ();
        return true;
    }


    public void rotatexaxis(int delta)
    {
        double xrot = delta * 2 * Math.PI / size().height;
        deltamat.postRotateX (xrot);
        repaint();
        return;
    }
    
    public void rotateyaxis(int delta)
    {
        double yrot = delta * 2 * Math.PI / size().width;
        deltamat.postRotateY (yrot);
        repaint();
        return;
    }
    
    public void onWheelMotion (Wheel wheel, int delta) {
        if (wheel.getOrientation () == Wheel.VERTICAL) {
            double xrot = delta * 2 * Math.PI / size().height;
            deltamat.postRotateX (xrot);
            }
        else {
            double yrot = delta * 2 * Math.PI / size().width;
            deltamat.postRotateY (yrot);
            }

        repaint ();
    }



    public void onWheelReset (Wheel wheel) {
        deltamat.unify ();
        repaint ();
        return;
    }

    public void onWheelResetButton()
    	{
			deltamat.unify();
			repaint();
			return;
		}



    public void setXRatio (float value)   {xratio = value;}
    public void setYRatio (float value)   {yratio = value;}
    public void setZRatio (float value)   {zratio = value;}




    public void onCheckerClick (Checker checker, boolean checked) {
        if ("transparency".equals (checker.text))
            opaqueplane = ! checked;

        if ("grid".equals (checker.text))
            needgrid = checked;

        repaint ();
        return;
    }

	public void onCheckrClick(boolean cgrid, boolean ctrans)
	{
		needgrid = cgrid;
		opaqueplane = ctrans;
		repaint();
		return;
	}

    public void onButtonClick (ImageButton button) {
        if ("xview".equals (button.text)) {
            viewpoint = new Vector3D (1.0f, 0.0f, 0.0f);
            setView ();
            deltamat.unify ();
            }

        if ("yview".equals (button.text)) {
            viewpoint = new Vector3D (0.0f, 1.0f, 0.0f);
            setView ();
            deltamat.unify ();
            }

        if ("zview".equals (button.text)) {
            viewpoint = new Vector3D (0.0f, 0.0f, 1.0f);
            setView ();
            deltamat.unify ();
            }

        repaint ();
        return;
    }





    public void startAnimation (int pathcount, int [] path, int aindex) {
        if (pathcount < 2)
            return;
        if (animode)
            return;

        this.pathcount = pathcount;
        this.path = path;
        animatex = (aindex == mapper.x);
        animatey = (aindex == mapper.y);
        animatez = (aindex == mapper.z);
        animatesize = (aindex == mapper.size);
        backupdata = new PlotData [datacount];
        animode = true;
    }


    public void stopAnimation () {
        animode = false;
        repaint ();
    }


    public void refresh (int curstep, float alpha) {
        this.curstep = curstep;
        this.alpha   = alpha;
        repaint ();
        }
}
