import java.awt.*;



class IconMap {
    private int x;
    private int y;
    private int width;
    private int height;


    IconMap (int posx, int posy, int w, int h) {
        x  = posx;
        y  = posy;
        width  = w;
        height = h;
    }


    void draw (Graphics graph, Image image, int posx, int posy) {
        Graphics clipgraph = graph.create (posx, posy, width, height);
        clipgraph.drawImage (image, -x, -y, null);
        return;
    }
}



public class Wheel extends Canvas {

    private Image       wheelimage;
    private int         size;
    private int         motionsize;
    private IconMap []  wheelmap;
    private int         orientation;
    private int         state;
    private boolean     buttonup;
    private Image       offscreen;
    private Graphics    offgraphics;
    private WheelClient client;



    public static final int VERTICAL    = 1;
    public static final int HORIZONTAL  = 2;




    private static int startoffset [] = {0, 2, 2, 5};
    private static int endoffset   [] = {6, 4, 2, 1};


    private static final IconMap verticalmap [] = {
        new IconMap (0,  0,  20, 30),
        new IconMap (0,  30, 20, 29),
        new IconMap (20, 0,  20, 30),
        new IconMap (20, 30, 20, 29),
        new IconMap (40, 0,  20, 30),
        new IconMap (40, 30, 20, 29),
        new IconMap (60, 0,  20, 30),
        new IconMap (60, 30, 20, 29)};


    private static final IconMap horizontalmap [] = {
        new IconMap (0, 139, 30, 20),
        new IconMap (30,139, 29, 20),
        new IconMap (0, 119, 30, 20),
        new IconMap (30,119, 29, 20),
        new IconMap (0,  99, 30, 20),
        new IconMap (30, 99, 29, 20),
        new IconMap (0,  79, 30, 20),
        new IconMap (30, 79, 29, 20)};


    private static final IconMap upmap   = new IconMap ( 0, 59, 20, 20);
    private static final IconMap downmap = new IconMap (20, 59, 20, 20);




    Wheel (int orient, int wheelsize, WheelClient wclient) {
        super ();

        client = wclient;
        orientation = orient;
        size = wheelsize;
        motionsize = size - 79;
        state = 0;
        buttonup = true;
        offscreen = null;
//        wheelimage = CHART3Dframe.ireader.getimage ("wheel.gif");

        if (orientation == VERTICAL) {
            wheelmap = verticalmap;
            resize (20, size);
            }
        else {
            wheelmap = horizontalmap;
            resize (size, 20);
            }
    }





    public int getOrientation () {
        return orientation;
    }



    public void paint (Graphics graph) {
        update (graph);
        return;
    }



    public synchronized void update (Graphics graph) {
        if (offscreen == null) {
            offscreen   = createImage (size().width, size().height);
            offgraphics = offscreen.getGraphics();
            }

        if (orientation == VERTICAL) {
            wheelmap [state*2].draw (offgraphics, wheelimage, 0, 0);
            wheelmap [state*2+1].draw (offgraphics, wheelimage, 0, 30+motionsize);
            if (buttonup)
                upmap.draw (offgraphics, wheelimage, 0, size-20);
            else
                downmap.draw (offgraphics, wheelimage, 0, size-20);


            offgraphics.setColor (Color.lightGray);
            offgraphics.fillRect (2, 30, 17, motionsize);
            offgraphics.setColor (Color.darkGray);
            offgraphics.drawLine (19, 30, 19, 29+motionsize);
            offgraphics.drawLine (4,  30,  4, 29+motionsize);
            offgraphics.setColor (Color.gray);
            offgraphics.drawLine (18, 30, 18, 29+motionsize);
            offgraphics.drawLine (0,  30,  0, 29+motionsize);
            offgraphics.setColor (Color.white);
            offgraphics.drawLine (1,  30,  1, 29+motionsize);
            offgraphics.drawLine (5,  29+motionsize/6,  5, 29+5*motionsize/6);
            offgraphics.drawLine (14, 29+motionsize/5, 14, 29+5*motionsize/6);
            offgraphics.setColor (Color.black);
            offgraphics.drawLine (15, 30, 15, 29+motionsize);
            }
        else  {
            wheelmap [state*2].draw (offgraphics, wheelimage, 0, 0);
            wheelmap [state*2+1].draw (offgraphics, wheelimage, 30+motionsize, 0);
            if (buttonup)
                upmap.draw (offgraphics, wheelimage, size-20, 0);
            else
                downmap.draw (offgraphics, wheelimage, size-20, 0);

            offgraphics.setColor (Color.lightGray);
            offgraphics.fillRect (30, 2, motionsize, 17);
            offgraphics.setColor (Color.darkGray);
            offgraphics.drawLine (30, 19, 29+motionsize, 19);
            offgraphics.drawLine (30, 4,  29+motionsize, 4);
            offgraphics.setColor (Color.gray);
            offgraphics.drawLine (30, 18, 29+motionsize, 18);
            offgraphics.drawLine (30,  0, 29+motionsize, 0);
            offgraphics.setColor (Color.white);
            offgraphics.drawLine (30,  1, 29+motionsize, 1);
            offgraphics.drawLine (29+motionsize/6,  5, 29+5*motionsize/6, 5);
            offgraphics.drawLine (29+motionsize/5, 14, 29+5*motionsize/6, 14);
            offgraphics.setColor (Color.black);
            offgraphics.drawLine (30, 15, 29+motionsize, 15);
            }


        int openpos  = startoffset [state];
        int closepos = motionsize - endoffset [state];
        int stepsize = 7;
        int span = motionsize;
        int spanacc = 0;
        int i;

        while (closepos > openpos && closepos-openpos >= stepsize) {
            if (orientation == VERTICAL) {
                offgraphics.setColor (Color.black);
                offgraphics.drawLine (6, openpos+30, 13, openpos+30);
                offgraphics.drawLine (6, closepos+30,   13, closepos+30);
                offgraphics.setColor (Color.white);
                for (i=0; i<stepsize-6; i++) {
                    offgraphics.drawLine (6, openpos+29-i, 13, openpos+29-i);
                    offgraphics.drawLine (6, closepos+29-i, 13, closepos+29-i);
                    }
                }
            else {
                offgraphics.setColor (Color.black);
                offgraphics.drawLine (openpos+30, 6, openpos+30, 13);
                offgraphics.drawLine (closepos+30, 6, closepos+30, 13);
                offgraphics.setColor (Color.white);
                for (i=0; i<stepsize-6; i++) {
                    offgraphics.drawLine (openpos+29-i, 6, openpos+29-i, 13);
                    offgraphics.drawLine (closepos+29-i, 6, closepos+29-i, 13);
                    }
                }

            spanacc += stepsize * 2;
            if (spanacc > span/2) {
                spanacc = 0;
                span = span/2;
                stepsize ++;
                }

            openpos  += stepsize;
            closepos -= stepsize;
            }



        graph.drawImage (offscreen, 0, 0, null);
    }



    int      prevx;
    int      prevy;
    boolean  mousedown = false;


    public boolean mouseDown (Event event, int x, int y) {
        boolean inbutton;
        if (orientation == VERTICAL)
            inbutton = (y > size-20);
        else
            inbutton = (x > size-20);

        if (inbutton) {
            buttonup = false;
            repaint ();
            return true;
            }
        else {
            prevx = x;
            prevy = y;
            mousedown = true;
            return true;
            }
    }


    public boolean mouseUp (Event event, int x, int y) {
        mousedown = false;
        if (! buttonup) {
            buttonup = true;
            repaint ();

            boolean inbutton;
            if (orientation == VERTICAL)
                inbutton = (y > size-20);
            else
                inbutton = (x > size-20);
            if (inbutton)
                client.onWheelReset (this);
            }

        return true;
    }



    public boolean mouseDrag (Event event, int x, int y) {
        if (mousedown == false)
            return true;

        int delta;
        if (orientation == VERTICAL) {
            delta = y - prevy;
            if (y > prevy)
                state = (state + 1) % 4;
            else
                state = (state + 3) % 4;
            }
        else {
            delta = x - prevx;
            if (x > prevx)
                state = (state + 1) % 4;
            else
                state = (state + 3) % 4;
            }

        client.onWheelMotion (this, delta);
        prevx = x;
        prevy = y;
        repaint ();
        return true;
    }

}






